from numpy import ndarray, \
                  array, \
                  newaxis, \
                  float64, \
                  vectorize, \
                  rollaxis, \
                  indices
from numpy.ma import sum

from permm.SpeciesGroup import Species
from warnings import warn
import operator
import re

ReactionGroup = str

__all__ = ['Stoic', 'ReactionFromString', 'Reaction']


class Stoic(ndarray):
    """
    Stoic is a sub-class of float with a role
    property.  The role property is 'r', 'p', or 'u'.
    r = reactant
    p = product
    u = undetermined
    
    Undetermined is useful when a species is part of a 
    net reaction and may play either role
    
    Stoic are created by
        Stoic(2., 'r')
    
    Stoic supports:
        __mul__: Stoic*(Stoic|Number)
        __rmul__: (Stoic|Number)*Stoic
        __add__: Stoic+(Stoic|Number)
    """
    __array_priority__ = 1001.
    def __new__(subtype, f, role):
        result = array(f).view(subtype)
        result.role = role
        return result
    
    def __array_wrap__(self, obj, context = None):
        result = obj.view(Stoic)
        if context is not None:
            args = context[1]
            if len(args) > 1:
                lhs, rhs = args[:2]
                if hasattr(lhs, 'role'):
                    if hasattr(rhs, 'role'):
                        if lhs.role != rhs.role:
                            result.role = 'u'
                        else:
                            result.role = lhs.role
                    else:
                            result.role = lhs.role                        
                else:
                    result.role = 'u'

        return result
        
    def __repr__(self):
        return "Stoic(%s, role = '%s')" % (str(self.asarray()), getattr(self, 'role', 'None'))
    
    def __str__(self):
        return self.__repr__()
            
    def asarray(self):
        return self.view(ndarray)
        

StoicArray = vectorize(Stoic, otypes = [object])
def AddRole(s1, s2):
    if isinstance(s1,Stoic):
        return s1
    else:
        return Stoic(s1, s2.role)
        
def StoicAdd(s1, s2):
    return Stoic.__add__(AddRole(s1,s2), s2)

StoicAdd = vectorize(StoicAdd, otypes = [object])

def ReactionFromString(rxn_str):
    """
    ReactionFromString is a convenience function.  It creates
    Reaction objects from a string with the following pattern:
        "(?P<reactants>.*)=(?P<rxn_type>[kj])[>]\s*(?P<products>.*)"
    where each reactant or product matches the following pattern:
        "(\s?(?P<sign>[+-])?\s?)?((?P<stoic>\d(\.(\d{1,3}(E\d{2})?)?)?)\*)?(?P<name>[A-Z]\w{0,3})\s?"

    For example:
        OH + OLE =k> 0.8*FORM + 0.33*ALD2 + 0.62*ALDX + 0.8*XO2 + 0.95*HO2 - 0.7 PAR
    """
    stoics = {}
    roles = {}
    species = ()
    
    reaction_re = re.compile("(?P<reactants>.*)->\[(?P<rxn_type>[kj])\]\s*(?P<products>.*)")

    species_re = re.compile("(\s?(?P<sign>[+-])?\s?)?((?P<stoic>\d{0,1}(\.(\d{1,3}(E\d{2})?)?)?)\*)?(?P<name>[a-zA-Z]\w*)(?:[ +=]|$)+",re.M)
    
    reaction_match = reaction_re.match(rxn_str)
    if reaction_match is None:
        raise SyntaxError, "Reactions must match the following patter\n\n<%%d> stoic*spc ([+-] stoic*spc)* =[kju]> [stoic*spc] ([+-] stoic*spc)*\n\n%s" % (rxn_str,)

    reactants = reaction_match.groupdict()['reactants']
    reaction_type = reaction_match.groupdict()['rxn_type']
    products = reaction_match.groupdict()['products']

    for spc in species_re.finditer(reactants):
        spc_g = spc.groupdict()
        name = spc_g['name']
        sign = spc_g['sign']
        stoic = spc_g['stoic']
        if sign is None:
            sign = ''

        if stoic is None:
            stoic = '1'

        if name in species:
            stoics[name] += -float(sign + stoic)
        else:
            species += (name,)
            stoics[name] = -float(sign + stoic)

        roles[name] = 'r'

    
    for spc in species_re.finditer(products):
        spc_g = spc.groupdict()
        name = spc_g['name']
        sign = spc_g['sign']
        stoic = spc_g['stoic']
        if sign is None:
            sign = '+'

        if stoic is None:
            stoic = '1'
        
        value = float(sign + stoic)
        if name in species:
            stoics[name] += value
            role = roles[name]
            if role == 'r':
                role = 'u'
            roles[name] = role
        else:
            stoics[name] = value
            roles[name] = 'p'
            species += (name,)
        
    return Reaction(reaction_type = reaction_type, roles = roles, **stoics)

class Reaction(object):
    """
    Reaction is an object that represents reaction groups.  The simplest
    case being a "single reaction" reaction group.
    
    Reaction groups support the following interfaces

        1) indexing (__getitem__) for stoiciometry
        2) multiplication (__mul__ and __rmul__) by arrays and numbers
        3) addition (__add__) of reactions or species
    
    A Reaction can also determine when a species is a reactant, product 
    or unspecified.  When a species is only present as a one role or the other 
    (reactant or product), it is always that role.  When a species is present 
    in the group as both, its current role is determined by the stoichiometry.
    
    Convenience functions:
        has_spc = spc in Reaction.species()
        has_rct = spc in Reaction.reactants()
        has_prd = spc in Reaction.products()
    
    Other functions:
        add_rct_spc
        add_prd_spc
    """
    __array_priority__ = 1000.
    def __init__(self, reaction_type = 'k', roles = {}, **stoic):
        """
            roles - dictionary of specific roles; for species whose 
                    stoichiometric sign is inconsistent with their 
                    role (i.e. X + Y =k> Z - .6*PAR)
            reaction_type - 'k' is kinetic, 'j' is photolysis, 'n' is net
            stoic - stoichiometry values provided as keywords by species; values
                    can be scalars or ndarrays
        """
        self.reaction_type = reaction_type
        self._species = tuple([k for k in stoic.keys()])
        self._roles = roles.copy()
        self._stoic = stoic.copy()
        
        try:
            self.shape = self._stoic[self._species[0]].shape
        except AttributeError, (e):
            self.shape = ()
            for spc in self._species:
                self._stoic[spc] = float64(self._stoic[spc])
            
        
        for k in self._species:    
            if k not in self._roles:
                self._roles[k] = {True: 'r', False: 'p'}[sum(stoic[k]) < 0]
        self._update_roles()

    def _update_roles(self):
        """
        Create static copy of species roles
        """
        self._reactants = ()
        self._products = ()
        self._unspecified = ()
        
        for k, v in self._roles.iteritems():
            if v == 'r':
                self._reactants += (k,)
            elif v == 'p':
                self._products += (k,)
            elif v == 'u':
                self._unspecified += (k,)
                if sum(self._stoic[k]) > 0:
                    self._products += (k,)
                else:
                    self._reactants += (k,)
    
    def __contains__(self, lhs):
        """
        Test if reaction has a species
        """
        if isinstance(lhs, Species):
            return len(set(lhs.names()).intersection(self._species)) > 0
        elif isinstance(lhs, str):
            return lhs in self._species
        else:
            raise TypeError, 'Unknown comparison: __contains__ for Reaction and %s' % str(type(lhs))
            
    def __getitem__(self, item):
        """
        Return a stoichiometry for a species or, if item is not a species, return
        a new reaction where stoichiometry is subset using item
        """
        if isinstance(item, Species):
            try:
                return self._stoic[item.name]
            except KeyError, (e):
                return self.get_spc(item)
        elif isinstance(item, str):
            return Stoic(self._stoic[item], role = self._roles[item])
        else:
            return Reaction(roles = self._roles, reaction_type = self.reaction_type, **dict([(k,v[item]) for k, v in self._stoic.iteritems()]))
    
    def __str__(self):
        """
        Report all values followed by the sum value for the reaction
        """
        result = ''
        temp = self.copy()
        temp.shape = ()
        if self.shape != ():
            result += '%d Reactions with shape %s\n' % (reduce(int.__mul__, self.shape), str(self.shape))
            for idx in indices(self.shape).reshape(len(self.shape), -1).swapaxes(0,1):
                idx = tuple(idx)
                result += str(idx) + ': '
                for spc in self._species:
                    temp._stoic[spc] = self._stoic[spc][idx]
                temp._update_roles()
                result += str(temp)+', \n'
            result = result[:-3]+'\n'
        sum_result = self.display(digits = 5, nspc = 1000)

        if result != '':
            result += '-' * (len(sum_result)+5) + '\nSum: '
        result += sum_result
            
        return result

    def __repr__(self):
        """
        Representation is same as string
        """
        return self.__str__()

    def display(self, digits = 5, nspc = 3, formatter = 'g'):
        reactants = [(self._stoic[rct].sum(), rct) for rct in self.reactants()]
        reactants.sort(reverse=False)

        products = [(self._stoic[prd].sum(), prd) for prd in self.products()]
        products.sort(reverse=True)
        if digits == None:
            str_temp = '%s'
            reactant_str = ' + '.join([str_temp % rct for stoic, rct in reactants][:nspc])
            product_str = ' + '.join([str_temp % prd for stoic, prd in products][:nspc])
        elif digits == -1:
            reactant_str = ' + '.join(['*'.join([str(-stoic), str(rct)]) for stoic, rct in reactants][:nspc])
            product_str = ' + '.join(['*'.join([str(stoic), str(prd)]) for stoic, prd in products][:nspc])
        else:
            str_temp = '%%.%d%s*%%s' % (digits, formatter)
            reactant_str = ' + '.join([str_temp % (-stoic,rct) for stoic, rct in reactants][:nspc])
            product_str = ' + '.join([str_temp % (stoic,prd) for stoic, prd in products][:nspc])

        if len(reactants) > nspc:
            reactant_str += ' + ...'
        if len(products) > nspc:
            product_str += ' + ...'

        return '%s ->[%s] %s' % (reactant_str, self.reaction_type, product_str)
        
    def __add__(self,rhs):
        """
        Add reactions to make a net reaction or add species to an existing reaction.
        """
        if isinstance(rhs,Reaction):
            kwds = {}
            
            for spc in self._species:
                kwds[spc] = self._stoic[spc]
            
            for spc in rhs._species:
                if kwds.has_key(spc):
                    kwds[spc] = kwds[spc] + rhs._stoic[spc]
                else:
                    kwds[spc] = rhs._stoic[spc]
            kwds['roles'] = roles = {}
            for spc in set(self._species+rhs._species):
                new_role = ''.join(set(self._roles.get(spc,'') + rhs._roles.get(spc,'')))
                if len(new_role) == 1:
                    roles[spc] = new_role
                else:
                    roles[spc] = 'u'
            
            if self.reaction_type == rhs.reaction_type:
                kwds['reaction_type'] = self.reaction_type
            else:
                kwds['reaction_type'] = 'n'

        elif isinstance(rhs,Species):
            return self.__add_if_in_spclist(rhs,self._species)

        else:
            raise TypeError, "Currently, only reactions can be added together"
        
        return Reaction(**kwds)
        
    def __rmul__(self,irrs):
        return self.__mul__(irrs)
        
    def __mul__(self,irrs):
        species = self._species
        values = dict([(k,v*irrs) for k, v in self._stoic.iteritems()])
        result = Reaction(roles = self._roles, reaction_type = self.reaction_type, **values)
        return result

    def __add_if_in_spclist(self,rhs,spc_list):
        if not self.has_spc(rhs):
            raise KeyError, 'Reaction has no components of %s' % rhs.name
        elif rhs.name in self._species:
            warn('Already has %s' % rhs.name)
            return self.copy()
        elif rhs.exclude:
            raise ValueError, 'Exclude is not supported'
        result = self.copy()
        new_stoic = result[rhs]
        
        result._roles[rhs.name] = new_stoic.role
        result._species += (rhs.name,)
        result._stoic[rhs.name] = new_stoic.view(ndarray)
        result._update_roles()

        return result
            
    def copy(self):
        """
        Create a copy of the reaction such that stoichiometry 
        are not shared
        """
        return Reaction(roles = self._roles, reaction_type = self.reaction_type, **dict([(k, v.copy()) for k, v in self._stoic.iteritems()]))

    def sum(self, axis = None):
        """
        Sum stoichiometries and create a scalar reaction
        """
        result = self.copy()
        for spc in result._species:
            result._stoic[spc] = self._stoic[spc].sum(axis)
        result.shape = ()
        result._update_roles()
        return result
        
    def mean(self, axis = None):
        """
        Mean stoichiometries and create a scalar reaction
        """
        result = self.copy()
        for spc in result._species:
            result._stoic[spc] = self._stoic[spc].mean(axis)
        result.shape = ()
        result._update_roles()
        return result
        
    def reactants(self):
        """
        Report all species acting as reactants; includes negative
        stoichiometries for unspecified role species
        """
        return self._reactants
        
    def products(self):
        """
        Report all species acting as products; includes positive
        stoichiometries for unspecified role species
        """
        return self._products
        
    def species(self):
        """
        Report all species
        """
        return self._species

    def unspecified(self):
        """
        Report all species with unspecified roles
        """
        return self._unspecified

    def get(self, item, default = None):
        try:
            return self.__getitem__(item)
        except:
            return default

    def get_spc(self, item, role = 'rpu'):
        species_list = ()
        if 'r' in role:
            species_list = species_list + self._reactants
        if 'p' in role:
            species_list = species_list + self._products
        if 'u' in role:
            species_list = species_list + self._unspecified

        species = set(item.names()).intersection(species_list)
        if len(species) == 0:
            raise KeyError, "%s does not contain %s" % (str(self.sum()), str(item))
        
        value = 0.
        for spc in species:
            value = value + item[spc] * self._stoic[spc]
        first_spc_role = self._roles[species.pop()]
        same_role = all([first_spc_role == self._roles[spc] for spc in species])
        
        if same_role:
            role = first_spc_role
        else:
            role = 'u'
        return Stoic(value, role = role)
        
    def produces(self, item):
        try:
            return self.get_spc(item, role = 'pu')
        except KeyError:
            return 0 * self[(self._products+self._reactants)[0]]

    def consumes(self, item):
        try:
            return -self.get_spc(item, role = 'ru')
        except KeyError:
            return 0 * self[(self._reactants+self._products)[0]]
    
    def has_spc(self,spc_grp):
        return spc_in_list(spc_grp,self._species)
        
    def has_rct(self,spc_grp):
        return spc_in_list(spc_grp,self._reactants)
        
    def has_prd(self,spc_grp):
        return spc_in_list(spc_grp,self._products)
    
    def add_rct_spc(self,rhs):
        return self.__add_if_in_spclist(rhs,self._reactants)
        
        
    def add_prd_spc(self,rhs):
        return self.__add_if_in_spclist(rhs,self._products)

def spc_in_list(spc_grp,local_list):
    if spc_grp.exclude:
        return not spc_in_list(-spc_grp, local_list)
    else:
        for s in spc_grp.names():
            if s in local_list:
                return True
        else:
            return False

import unittest

class ReactionTestCase(unittest.TestCase):
    def setUp(self):
        self.spcs = dict(NO2 = Species(name = 'NO2', names = ['NO2'], stoic = [1.]),
                         NO = Species(name = 'NO', names = ['NO'], stoic = [1.]),
                         HNO3 = Species(name = 'HNO3', names = ['HNO3'], stoic = [1.]),
                         O = Species(name = 'O', names = ['O'], stoic = [1.]),
                         OH = Species(name = 'OH', names = ['OH'], stoic = [1.]),
                         HO2 = Species(name = 'HO2', names = ['HO2'], stoic = [1.]),
                         O3 = Species(name = 'O3', names = ['O3'], stoic = [1.]),
                         PAR = Species(name = 'PAR', names = ['PAR'], stoic = [1.]),
                         NTR = Species(name = 'NTR', names = ['NTR'], stoic = [1.]),
                         FORM = Species(name = 'FORM', names = ['FORM'], stoic = [1.]),
                         ALD2 = Species(name = 'ALD2', names = ['ALD2'], stoic = [1.]),
                         ALDX = Species(name = 'ALDX', names = ['ALDX'], stoic = [1.]),
                         )
        exec('NOx = NO2 + NO', globals(), self.spcs)
        exec('ALD = ALDX + ALD2 + FORM', globals(), self.spcs)
        rxn_strings = {'NO2hv': 'NO2 ->[j] NO + O',
                       'OplO3': 'O + O2 + M ->[k] O3 + M',
                       'NTRplOH': 'NTR + OH ->[k] HNO3 + HO2 + 0.330*FORM + 0.330*ALD2 + 0.330*ALDX - 0.660*PAR'
                      }
        self.rxns = {}
        for label, rxn_str in rxn_strings.iteritems():
            self.rxns[label] = ReactionFromString(rxn_str)
        
    def testHasSpc(self):
        r1 = self.rxns['NO2hv']
        spcs = [self.spcs[k] for k in ['NO2', 'NO', 'NOx']]
        for spc in spcs:
            self.assertTrue(r1.has_spc(spc))
            
    def testHasRole(self):
        r1 = self.rxns['NO2hv']
        rs = self.spcs['NO2']
        ps = self.spcs['NO']
        self.assertTrue(r1.has_rct(rs))
        self.assertFalse(r1.has_rct(ps))
        self.assertFalse(r1.has_prd(rs))
        self.assertTrue(r1.has_prd(ps))
        
    def testConsumes(self):
        r1 = self.rxns['NO2hv']
        rs = self.spcs['NO2']
        ps = self.spcs['NO']
        ns = self.spcs['O3']
        self.assertTrue(r1.consumes(rs) == 1)
        self.assertTrue(r1.consumes(ps) == 0)
        self.assertTrue(r1.consumes(ns) == 0)
        
    def testProduces(self):
        r1 = self.rxns['NO2hv']
        rs = self.spcs['NO2']
        ps = self.spcs['NO']
        ns = self.spcs['O3']
        self.assertTrue(r1.produces(rs) == 0)
        self.assertTrue(r1.produces(ps) == 1)
        self.assertTrue(r1.produces(ns) == 0)
        
    def testGet(self):
        r1 = self.rxns['NO2hv']
        rs = self.spcs['NO2']
        self.assertTrue(r1[rs] == -1)
        
    def testReactants(self):
        r1 = self.rxns['NTRplOH']
        self.assertEquals(set(r1.reactants()), set(('OH', 'NTR')))
        
        
    def testProducts(self):
        r1 = self.rxns['NTRplOH']
        self.assertEquals(set(r1.products()), set(('PAR', 'FORM', 'ALD2', 'HNO3', 'HO2', 'ALDX')))
        
    def testMulScalar(self):
        r1 = self.rxns['NTRplOH']
        a = 5.83
        r2 = r1 * a

        for spcn in "NTR OH".split():
            spc = self.spcs[spcn]
            self.assertAlmostEqual(r2[spc], -a)

        for spcn in "HNO3 HO2".split():
            spc = self.spcs[spcn]
            self.assertAlmostEqual(r2[spc], a)

        for spcn in "FORM ALD2 ALDX".split():
            spc = self.spcs[spcn]
            self.assertAlmostEqual(r2[spc], a * .33)

        for spcn in ["ALD"]:
            spc = self.spcs[spcn]
            self.assertAlmostEqual(r2[spc], a * .99)

        for spcn in ["PAR"]:
            spc = self.spcs[spcn]
            self.assertAlmostEqual(r2[spc], -a * .66)
        
    def testMulArray(self):
        from numpy import arange, round
        a = arange(0, 60, dtype = 'd').reshape(3,4,5) + .3
        r1 = self.rxns['NTRplOH']
        r2 = r1 * a

        for spcn in "NTR OH".split():
            spc = self.spcs[spcn]
            self.assertTrue((r2[spc] == -a).all())

        for spcn in "HNO3 HO2".split():
            spc = self.spcs[spcn]
            self.assertTrue((r2[spc] == a).all())

        for spcn in "FORM ALD2 ALDX".split():
            spc = self.spcs[spcn]
            self.assertTrue((r2[spc] == (a * float64(.33))).all())

        for spcn in ["ALD"]:
            spc = self.spcs[spcn]
            # Value is not exact because of addition
            self.assertTrue((round(r2[spc], decimals = 8) == round(a * float64(.99), decimals = 8)).all())

        for spcn in ["PAR"]:
            spc = self.spcs[spcn]
            self.assertTrue((r2[spc] == (-a * .66)).all())
        
if __name__ == '__main__':
    unittest.main()


from Mechanism import Mechanism

def get_pure_mech(mechanism):
    from os.path import dirname, \
                        abspath, \
                        join, \
                        exists
    if exists(mechanism) and '.yaml' in mechanism:
        mech_path = mechanism
    else:
        mech_path = abspath(join(dirname(__file__),'mechanisms', mechanism,'%s.yaml' % (mechanism,)))
    
    if not exists(mech_path):
        raise ImportError, "Mechanism you supplied is not a known mechanism and is not a file path to a definition"
        
    return Mechanism(mech_path)
    
def get_prepared_mech(mechanism):
    from mechanisms.cb05_camx.mechprep import cb05_camx_prep as cb05_camx
    from mechanisms.cb05_cmaq.mechprep import cb05_cmaq_prep as cb05_cmaq
    from mechanisms.geos_chem.mechprep import geos_chem_prep as geos_chem
    mech_object = get_pure_mech(mechanism)
    mech_prep = eval(mechanism)

    return mech_prep(mech_object)


__all__ = ['cb05_camx', 'cb05_cmaq', 'geos_chem', 'saprc99_cmaq', 'saprc07_cmaq', 'racm2_cmaq', 'mz4_kpp', 'small_strato']
from permm.getmech import get_pure_mech
from yaml import load
from os.path import dirname, \
                    abspath, \
                    join
atoms_path = abspath(join(dirname(__file__), 'atoms.yaml'))

atoms = load(file(atoms_path).read())
for mech in __all__:
    exec('def %s():\n    return get_pure_mech("%s")' % (mech, mech), globals(), locals())



from numpy import array, zeros
def cycles(mech, species_group, indirect = False):
    mech.globalize(globals())
    species_group_names = species_group.names()
    species_group_cycles = {}
    for spc_name in species_group.names():
        spc = eval(spc_name)
        spc_names = set(spc.names())
        this_spc_cycles = species_group_cycles[spc] = {}
        init = make_net_rxn(-species_group, spc)[spc]
        this_spc_cycles['source'] = init
        prop_total = zeros(init.shape, dtype = 'f')
        for prop in get_irrs(species_group, spc):
            made = prop.get_spc(spc)
            prop_total += made
            rad_rcts = [eval(rct) for rct in set(prop.reactants()).intersection(species_group_names)]
            rad_rct_stoics = array([prop.get_spc(rad, 'r') for rad in rad_rcts])
            rad_rct_stoic = rad_rct_stoics.sum(0)
            rad_frac = rad_rct_stoics / rad_rct_stoic[None,:]
            for radi, rad in enumerate(rad_rcts):
                if (made != 0).any():
                    this_spc_cycles.setdefault('propfrom', {})[rad] = made / rad_frac[radi]

        sink_total = zeros(init.shape, dtype = 'f')
        template = zeros(init.shape, dtype = 'f')
        for sink in get_irrs(spc, -species_group):
            made = sink.get_spc(spc)
            sink_total += made
            nrad_rcts = [eval(rct) for rct in set(sink.reactants()).difference(spc_names)]
            if nrad_rcts == []:
                nrad_rcts = [spc]
            nrad_rct_stoics = array([sink.get_spc(rad, 'r') for rad in nrad_rcts])
            nrad_rct_stoic = nrad_rct_stoics.sum(0)
            nrad_frac = nrad_rct_stoics / nrad_rct_stoic[None,:]
            for nradi, nrad in enumerate(nrad_rcts):
                if (made != 0).any():
                    this_spc_cycles.setdefault('sinkby', {}).setdefault(nrad, template.copy())[:] = made / nrad_frac[nradi]
        
        this_spc_cycles['reacted'] = make_net_rxn(spc)[spc]
        this_spc_cycles['prop'] = prop_total.copy()
        this_spc_cycles['sink'] = sink_total.copy()

        exec('propeff = 1 - sink / reacted', globals(), this_spc_cycles)
        exec('total = source + prop', globals(), this_spc_cycles)
        
    if indirect:
        for spc in this_spc_cycles.keys():
            this_spc_cycles = species_group_cycles[spc]
            for prop_source, prop in this_spc_cycles['propby'].iteritems():
                propeff = species_group_cycles[prop_source]['propeff']
                
    return species_group_cycles
    


from utils import AttrDict
from collections import defaultdict
import operator
import yaml
import re
import sys
from numpy import * # Explicitly using dtype, array and ndarray; providing all default numpy to __call__interface
from warnings import warn
from SpeciesGroup import Species, species_sum
from ProcessGroup import Process
from ReactionGroup import ReactionFromString
from IPRArray import Processes_ProcDelimSpcDict
from graphing.timeseries import irr_plot, phy_plot, plot as tplot
from Shell import load_environ
from PseudoNetCDF.sci_var import PseudoNetCDFVariable
from netcdf import NetCDFVariable

__all__ = ['Mechanism']

_spc_def_re = re.compile(r'(?P<stoic>[-+]?[0-9]*\.?[0-9]+)(?P<atom>\S+)(?=\s*\+\s*)?')
_numre = re.compile('(\d+)')

class Mechanism(object):
    """
    The mechanism object is the central repository for information
    about a chemcial model.  It is aware of species, reactions, species
    groups (e.g. NOy), net reactions (e.g. PAN <-> C2O3 + NO2).  
    
    The mechanism object also has introspective tools.  In addition, it
    provides an interpreter through its call interface.
    
    Reaction data can be augmented with IRR data via set_irr or set_mrg
    """
    def __init__(self, yaml_path):
        """
        Initialization requires a yaml_path or a dictionary.
        The dictionary should have the following keys:
        
        species_list - a list of string names for chemical species
                        (i.e. O3)
        reaction_list - a list of strings representing reactions 
                        (see ReactionGroup.ReactionFromString)
        net_reaction_list - a dictionary of reaction additions 
                            (i.e. NAME: RXN_01 + RXN_02)
        process_group_list - a dictionary of process additions
                            (i.e. VertAdv = Top_Adv + Bottom_Adv)
        """
        import os
        if isinstance(yaml_path,str):
            if os.path.exists(yaml_path):
                yaml_file = yaml.load(file(yaml_path))
            else:
                yaml_file = yaml.load(yaml_path)
        elif isinstance(yaml_path,dict):
            yaml_file = yaml_path
        
        yaml_file = AttrDict(yaml_file)
        self.__yaml_file = yaml_file
        self.mechanism_comment = yaml_file.get('comment', '')
        self.species_dict = AttrDict()
        if yaml_file.has_key('species_list'):
            for spc, spc_def in yaml_file.species_list.iteritems():
                if spc_def == 'IGNORE':
                    self.species_dict[spc] = Species(name = spc, names = [spc], stoic = [1])
                else:
                    spc_atoms = dict([(atom, eval(stoic)) for stoic, atom in _spc_def_re.findall(spc_def)])
                    self.species_dict[spc] = Species(name = spc, names = [spc], stoic = [1], atom_dict = {spc: spc_atoms})
                        
        self.reaction_dict = AttrDict()
        reaction_species = []
        for rxn_name, rxn_str in yaml_file.reaction_list.iteritems():
            rxn = self.reaction_dict[rxn_name] = ReactionFromString(rxn_str)
            reaction_species += rxn.species()
        
        reaction_species = set(reaction_species)
        for spc in [spc for spc in reaction_species if not self.species_dict.has_key(spc)]:
            self.species_dict[spc] = Species(name = spc, names = [spc], stoic = [1])

        for spc_grp_def in yaml_file.get('species_group_list',[]):
            grp_name = spc_grp_def.split('=')[0].strip()
            if (spc_grp_def.count('+') + spc_grp_def.count('-')) > 4:
                spc_grp_def = '=species_sum(['.join(spc_grp_def.replace('+',',').replace('-',',').split('='))+'])'
            exec(spc_grp_def, None, self.species_dict)
            self.species_dict[grp_name].name = grp_name

        self.net_reaction_dict = yaml_file.get('net_reaction_list',{})
        self.variables = {}
        load_environ(self, self.variables)
            
    def __call__(self, expr, env = globals()):
        """
        Evaluate string expression in the context of mechanism
        species, species groups, reactions, net reactions and processes
        """
        return eval(expr, env, self.variables)
    
    def __getitem__(self,item):
        """
        Provide a single getitem interface for mechanism species, species 
        groups, reactions, net reactions and processes
        """
        return self.variables[item]
        
    def __add_spc_to_reactions(self,rxn_list, spc):
        """
        Add a species to a reaction as an accumulation of components
          if RXN1 = CH3C(O)CH3 -j> CH3C(O)O2 + CH3O2
          and SPC = Species(CH3C(O)O2 + CH3O2, name = 'AO2')
          
          mech.__add_spc_to_reactions([RXN1],SPC)
          
          would make RXN1 = CH3C(O)CH3 -j> CH3C(O)O2 + CH3O2 + 2*AO2
        
        For more detail, see ReactionGroup.Reaction.__add__ interface
        """
        if not self.species_dict.has_key(spc.name):
            self.species_dict[spc.name] = spc
            
        for rxn in rxn_list:
            self.reaction_dict[rxn] = self.reaction_dict[rxn] + spc
        
        return len(rxn_list)
    
    def __ensure_species(self,spc):
        """
        Return species (created if necessary)
        """
        if isinstance(spc,str):
            spc = self.species_dict[spc]
        return spc
        
    def add_rct_to_reactions(self, spc):
        """
        Add spc to reactions where components are reactants
        see __ensure_species
        """
        spc = self.__ensure_species(spc)
        
        rxns = self.find_rxns(spc,[])
        return self.__add_spc_to_reactions(rxns, spc)

    def add_prd_to_reactions(self, spc):
        """
        Add spc to reactions where components are products
        see __ensure_species
        """
        spc = self.__ensure_species(spc)
        
        rxns = self.find_rxns([],spc)
        return self.__add_spc_to_reactions(rxns, spc)

    def add_spc_to_reactions(self, spc):
        """
        Add spc to reactions where components are reactants
        or products see __ensure_species
        """
        nrcts = self.add_rct_to_reactions(spc)
        nprds = self.add_prd_to_reactions(spc)
        return nrcts+nprds

    def add_spc_to_reactions_with_condition(self, spc, condition):
        """
        spc - must be a species object or string
        condition - must be a function that, given a reaction, evaluates to true or false
        """
        spc = self.__ensure_species(spc)
        
        new_rxn_def = {}
        for rxn_name, rxn in self.reaction_dict.iteritems():
            if condition(rxn):
                rxn = rxn + spc
                new_rxn_def[rxn_name] = rxn
        
        self.reaction_dict.update(new_rxn_def)

        return len(new_rxn_def)
    
    def get_rxns(self, reactants = [], products =[], logical_and = True, reaction_type = None):
        """
        Get reaction objects that meet the criteria specified by reactants, products and logical_and.
        
        reactants - filter mechanism reactions for reactions with all reactant(s)
        products - filter mechanism reactions for reactions with all product(s)
        logical_and - a boolean indicating how to combine reactant and product filters
                      True: reaction is in both filters (i.e. reactants AND products)
                      False: reaction is in either filter (i.e. reactants OR products)
        """
        return [self.reaction_dict[rk] for rk in self.find_rxns(reactants = reactants, products = products, logical_and = logical_and, reaction_type = reaction_type)]

    def get_irrs(self, reactants = [], products =[], logical_and = True, reaction_type = None):
        """
        Get reaction objects that meet the criteria specified by reactants, products and logical_and.
        
        reactants - filter mechanism reactions for reactions with all reactant(s)
        products - filter mechanism reactions for reactions with all product(s)
        logical_and - a boolean indicating how to combine reactant and product filters
                      True: reaction is in both filters (i.e. reactants AND products)
                      False: reaction is in either filter (i.e. reactants OR products)
        """
        return [self(rk) for rk in self.find_rxns(reactants = reactants, products = products, logical_and = logical_and, reaction_type = reaction_type)]

    def find_rxns(self, reactants = [], products =[], logical_and = True, reaction_type = None):
        """
        Get reaction names that meet the criteria specified by reactants, products and logical_and.
        
        reactants - filter mechanism reactions for reactions with all reactant(s)
        products - filter mechanism reactions for reactions with all product(s)
        logical_and - a boolean indicating how to combine reactant and product filters
                      True: reaction is in both filters (i.e. reactants AND products)
                      False: reaction is in either filter (i.e. reactants OR products)
        """
        if isinstance(reactants, (Species, str)):
            reactants = [reactants]
        
        reactants = [self.__ensure_species(spc) for spc in reactants]
            
        if isinstance(products, (Species, str)):
            products = [products]

        products = [self.__ensure_species(spc) for spc in products]
            
        result = [(rxn_name, rxn) for rxn_name, rxn in self.reaction_dict.iteritems()]


        if reactants != []:
            reactant_result = [rxn_name for rxn_name, rxn in result if all([rxn.has_rct(rct) for rct in reactants])]
        else:
            reactant_result = [rn for rn, rv in result]
        
        if products != []:
            product_result = [rxn_name for rxn_name, rxn in result if all([rxn.has_prd(prd) for prd in products])]
        else:
            product_result = [rn for rn, rv in result]
        
            
        if logical_and:
            reaction_list = set(reactant_result).intersection(product_result)
        else:
            reaction_list = set(reactant_result+product_result)

        if reaction_type is not None:
            reaction_result = [rn for rn, rv in result if rv.reaction_type in reaction_type]
            reaction_list = reaction_list.intersection(reaction_result)

        
        result = list(reaction_list)
        result.sort()
        
        return result
    
    def yaml_net_rxn(self, rxns):
        """
        Create the YAML representation of a net reaction for the supplied
        reactions
        """
        species = list(set(reduce(operator.add, [self.reaction_dict[rxn].species() for rxn in rxns])))
        
        for role in ('r','p'):
            for spc in species:
                for rxn in rxns:
                    method = dict( r = 'has_rct', p = 'has_prd')
                    if getattr(self.reaction_dict[rxn], method)(spc):
                        pass
    
    def subst_net_rxn(self, reactants = [], products = [], logical_and = True, reaction_type = None, name = None):
        rxns = self.find_rxns(reactants = reactants, products = products, logical_and = logical_and, reaction_type = reaction_type)
        if len(rxns) > 1:
            eval_str = ' + '.join(rxns)
            if name is None:
                name = eval_str
            nrxn = eval(eval_str, globals(), self.irr_dict)
            for rxn in rxns:
                del self.irr_dict[rxn]
                del self.reaction_dict[rxn]
            self.irr_dict[name] = nrxn
            self.reaction_dict[name] = nrxn.sum()
            load_environ(self, self.variables)
        
        
    def make_net_rxn(self, reactants = [], products = [], logical_and = True, reaction_type = None):
        """
        Sum each reaction in find_rxns(reactants, procucts, logical_and)
        and return a net reaction
        
        for more information see find_rxns
        """
        rxns = self.find_rxns(reactants, products, logical_and, reaction_type)
        result = self(' + '.join(rxns))
        
        return result

    def print_net_rxn(self, reactants = [], products = [], logical_and = True, reaction_type = None):
        """
        Sum each reaction in find_rxns(reactants, procucts, logical_and)
        and return a net reaction
        
        for more information see find_rxns
        """
        net_rxn = self.make_net_rxn(reactants, products, logical_and, reaction_type)
        
        print net_rxn

    def print_rxns(self, reactants = [], products = [], logical_and = True, reaction_type = None, sortby = None, reverse = False, digits = -1, nspc = 10000):
        """
        For each reaction in find_rxns(reactants, procucts, logical_and),
        print the reaction
        
        for more information see find_rxns
        """
        rxns = self.find_rxns(reactants, products, logical_and, reaction_type)

        rxns = [(rxn, self.reaction_dict[rxn]) for rxn in rxns]
        try:
            if sortby is None:
                sortby = ([p for p in _ensure_list(products) if not p.exclude]+[r for r in _ensure_list(reactants)  if not r.exclude])[0]
            rxns = [(rxno[sortby], rxn, rxno) for rxn, rxno in rxns]
            rxns.sort(reverse = reverse)
        except:
            rxns = [(_rxn_ordinal(rxn), rxn, rxno) for rxn, rxno in rxns]
            rxns.sort()
            warn("Not all reactions contain %s; check query and/or explicitly define sortby species" % str(sortby))

        for mass, rxn, rxno in rxns:
            print rxn, rxno.display(nspc = nspc, digits = digits)

    def plot_rxn_list(self, reactions, plot_spc = None, path = None, show = False, **kwds):
        """
        Returns:
            a matplotlib figure with a line for each reaction indexed by plot_spc
        
        Parameters:
            reactions - list of reaction objects
            plot_spc - species to plot
            path - path to savefigure
            show - try to display figure interactively

        Other keywords:
            accepts any keyword accepted by permm.graphing.timeseries.irr_plot
        """
        if plot_spc is None:
            plot_spc = self((reactions[0][0].products() + reactions[0][0].reactants())[0])
            
        fig = irr_plot(self, reactions = reactions, species = plot_spc, **kwds)
        if path is not None and show:
            fig.savefig(path)
        else:
            from pylab import show
            show()
        
        return fig

    def plot(self, y, path = None, stepped = True, end_date = True, time_slice = slice(None), figure_settings = {}, axis_settings = {}, line_settings = {}):
        """
        plot creates a timeseries plot of any numerical array.
        """
        fig = tplot(self, y, stepped = stepped, end_date = end_date,
                    figure_settings = figure_settings,
                    axis_settings = axis_settings,
                    line_settings = line_settings)
        if path is not None:
            fig.savefig(path)

        return fig

    def plot_rxns(self, reactants = [], products = [], logical_and = True, reaction_type = None,
                  plot_spc = None, combine = [()], nlines = 8, **kwds):
        """
        Creates figure of reactions
        
        Steps:
        1. Query reactions using get_rxns(reactants, procucts, logical_and)
        2. net reaction sets specified in combine (i.e. combine = [('IRR_1', 'IRR_2'), ...])
        3. sort queried reactions by absolute plot_spc change (+/-)
        3. combine all reactions that are not in the top (nlines - 1)
        4. plot the plot_spc change from each (nlines - 1) individual reactions
           and the 1 net reaction
        
        Returns:
            matplotlib figure with each queried reaction plotted for production/consumption
            of plot_spc.
        
        Parameters:
            reactants - instance or list of Species objects to require
            products - instance or list of Species objects to require
            logical_and - if true, require reactants and products; false, require reactants or products
            reaction_type - some combination of kjn: thermal (k); photolysis (j); net (n)
            plot_spc - Species instance to be plotted, defaults to first product or reactant
            combine - reaction keys to combine into net reactions
            nlines - maximum reactions to plot; selected based on maximum absolute plot_spc change
        
        Other keywords:
            permm.Mechanism.plot_rxn_list or 
            permm.graphing.timeseries.irr_plot
        
        """
        if plot_spc is None:
            plot_spc = (_ensure_list(products)+_ensure_list(reactants))[0]
        reactions = self.find_rxns(reactants, products, logical_and, reaction_type)
        
        if reactions == []:
            raise ValueError, "Your query didn't match any reactions; check your query and try again (try print_rxns)."

        reactions = [ rxn for rxn in reactions if rxn not in reduce(operator.add, combine) ]
        nlines = min(nlines, len(reactions)+1)
        if combine != [()]:
            reactions = reactions + map(lambda t2: '+'.join(t2), combine)
        
        reactions = [ (abs(self('(%s)' % (rxn))[plot_spc]).sum(),rxn) for rxn in reactions]
    
        reactions.sort(reverse = True)

        reactions = [r for v,r in reactions]
        for rxn in reactions[nlines-1:]:
            try:
                other = other + self('(%s)' % (rxn,))
            except:
                other = self('(%s)' % (rxn,))
        
        try:
            reactions = [self('(%s)' % (rxn, )) for rxn in reactions[:nlines-1]] + [other]
        except:
            reactions = [self('(%s)' % (rxn, )) for rxn in reactions[:nlines-1]]
        
        return self.plot_rxn_list(reactions = reactions, plot_spc = plot_spc, **kwds)
        
    def print_irrs(self, reactants = [], products = [], logical_and = True, reaction_type = None, factor = 1., sortby = None, reverse = False, slice = slice(None), nspc = 100000, digits = -1, formatter = 'g'):
        """
        For each reaction in find_rxns(reactants, procucts, logical_and),
        print the reaction summed for the entire timeseries.
        
        for more information see find_rxns
        """
        if not hasattr(self,'irr_dict'):
            raise ValueError, "Net reactions are only available when IRR has been loaded"
            
        rxns = self.find_rxns(reactants, products, logical_and, reaction_type)
        irrs = [(rxn, self.irr_dict[rxn][slice].sum()) for rxn in rxns]
        
        try:
            if sortby is None:
                sortby = ([p for p in _ensure_list(products) if not p.exclude]+[r for r in _ensure_list(reactants)  if not r.exclude])[0]
            irrs = [(irr[sortby], rxn, irr) for rxn, irr in irrs]
            irrs.sort(reverse = reverse)
        except:
            irrs = [(_rxn_ordinal(rxn), rxn, irr) for rxn, irr in irrs]
            irrs.sort()
            warn("Not all reactions contain %s; check query and/or explicitly define sortby species" % str(sortby))

        irrs = [(rxn, irr) for mass, rxn, irr in irrs]


        for rxn, irr in irrs:
            print rxn, (irr * factor).display(nspc = nspc, digits = digits)
        
    def set_mrg(self,mrg, use_net_rxns = True, use_irr = True, use_ipr = True):
        """
        Add process analysis from a 1D merged IRR/IPR file
        """
        self.mrg = mrg
        if use_irr:
            try:
                self.set_irr(mrg.variables['IRR'], mrg.Reactions.split(), use_net_rxns = use_net_rxns)
            except:
                self.set_irr()
        if use_ipr:
            try:
                self.set_ipr(mrg.variables['IPR'])
            except:
                self.set_ipr()
                        
        
        load_environ(self, self.variables)
        
    def set_irr(self,irr = None, ReactionNames = None, use_net_rxns = True):
        """
        Add process analysis from a 2D merged IRR array dim(TIME,RXN)
        """
        if not irr is None:
            irr_type = dtype(dict(names = ReactionNames, formats = irr[:].dtype.char*len(ReactionNames)))
                
            self.irr = irr[:].view(dtype = irr_type).squeeze().view(type = PseudoNetCDFVariable)
            self.irr.units = irr.units

        self.__use_net_rxns = use_net_rxns
        self.apply_irr()

    def apply_irr(self):
        self.irr_dict = {}
        for rxn_name, rxn in self.reaction_dict.iteritems():
            if hasattr(self, 'irr'):
                try:
                    self.irr_dict[rxn_name] = rxn * self.irr[rxn_name]
                except ValueError, (e):
                    self.irr_dict[rxn_name] = rxn * zeros(self.irr.shape, 'f')
                    warn("IRR does not contain %s: skipped." % rxn_name)
            else:
                try:
                    self.irr_dict[rxn_name] = rxn * self.mrg.variables[rxn_name][:].view(type = PseudoNetCDFVariable)
                except (KeyError, ValueError), (e):
                    warn("IRR does not contain %s: skipped." % rxn_name)
                
        if self.__use_net_rxns and len(self.irr_dict)>0:
            self.nreaction_dict = {}
            for nrxn_name, nrxn in self.net_reaction_dict.iteritems():
                try:
                    self.nreaction_dict[nrxn_name] = eval(nrxn, None, self.irr_dict)
                except Exception, (e):
                    warn("Predefined net rxn %s is not available; %s" % (nrxn_name, str(e)))

        load_environ(self, self.variables)
        
    def set_ipr(self, ipr = None, processes = None):
        """
        Add process analysis from a 3D merged IPR array (TIME,SPC,PROC)
        """
        if ipr is None:
            if hasattr(self.mrg, 'Processes'):
                processes = self.mrg.Processes.split()
                self.process_dict = Processes_ProcDelimSpcDict(processes, self.mrg.variables)
            else:
                warn("Unable to load IPR; missing Species and Processes attributes")
                return
        elif isinstance(ipr, dict):
            if processes is None:
                raise ValueError, "When ipr is a dictionary, processes must be provided as a list of process names"
            self.process_dict = Processes_ProcDelimSpcDict(processes, self.mrg.variables)
        elif isinstance(ipr, (PseudoNetCDFVariable, NetCDFVariable)):
            self.process_dict = {}
            for pi, prc in enumerate(processes):
                self.process_dict[prc] = Process(prc, default_unit = getattr(ipr, 'units', 'Unknown'), **dict([(spc, ipr[si, pi]) for si, spc in enumerate(species)]))
        else:
            return

        for prc_name, prc in self.__yaml_file.get('process_group_list', {}).iteritems():
            try:
                self.process_dict[prc_name] = eval(prc,{},self.process_dict)
                self.process_dict[prc_name].name = prc_name
            except:
                warn("Cannot create %s process group" % prc_name)
            
        # Add extra species for names in IPR
        spcs = set(reduce(list.__add__, [[spc for spc in proc.keys()] for proc in self.process_dict.values()]))
        new_spcs = spcs.difference(self.species_dict.keys())
        for name in new_spcs:
            if not self.species_dict.has_key(name):
                self.species_dict[name] = Species(name = name, names = [name], stoic = [1.])

    def add_rxn(self, rxn_key, rxn_str):
        self.reaction_dict[rxn_key] = ReactionFromString(rxn_str)
        if hasattr(self, 'irr_dict'):
            self.irr_dict[rxn_key] = ReactionFromString(rxn_str)
            self.irr_dict[rxn_key] *= self.irr[rxn_key]
        load_environ(self, self.variables)
    
    def plot_proc(self, species, path = None, **kwds):
        """
        species - permm.SpeciesGroup.Species object
        path - path for saved figure
        kwds - * title - title
               * init - Name of initial concentration process
               * final - Name of final concentration process
               * linestyle - default line style (default: '-')
               * linewidth - default line width (default: 3)
               * marker - default line marker style (default: None)
               * ncol - number of legend columns (default: 1)
               * fig - figure to plot on (default: None)
               * cmap - matplotlib color map for lines (default: None)
               * filter - remove processes with zero values (default: True)
               * <process name1> - process names from mech.process_dict can be 
                                   provided to limit the processes shown.  When 
                                   provided, a process should be a dictionary of 
                                   matplotlib plot options (common: linestyle,
                                   linewidth, label, marker).  The dictionary can
                                   be empty.
               * <process name2> - same as process 1
               * <process nameN> - same as process 1
               * end_date - times are for the time period end
               
               
        """
        fig = phy_plot(self, species, **kwds)
        if path is not None:
            fig.savefig(path)
        else:
            if kwds.get('show', True):
                from pylab import show
                show()
        return fig

    def globalize(self, env):
        load_environ(self, env)

def _ensure_list(x):
    if isinstance(x, Species):
        return [x]
    else:
        return x

def _rxn_ordinal(rxnlabel):
    result = _numre.search(rxnlabel)
    if result is None:
        return None
    else:
        val = result.groups()[0]
        try:
            return eval(val)
        except:
            return int(val)
        


from utils import AttrDict
__all__ = ['Process']

class Process(AttrDict):
    """
    Process is a group of processes where the simplest case is
    a 1 process group where names[0] == name
    
    The easiest way to create a process is with keywords:
        Process(
                name = '<process name>',
                names = ['component name1', ..., 'component name2']
               )
        
    Processes can be added together to create larger processes
        >>> A_W
        {'exclude': False, 'names': ['A_W'], 'name': 'A_W'}
        >>> A_E
        {'exclude': False, 'names': ['A_E'], 'name': 'A_E'}
        >>> H_Trans = A_W + A_E + A_S + A_N + D_S + D_W + D_N + D_E
        >>> H_Trans.name = 'H_Trans'
        >>> H_Trans
        {
            'name': 'H_Trans',
            'names': ['A_W', 'A_E', 'A_S', 'A_N', 'D_S', 'D_W', 'D_N', 'D_E'],
            'exclude': False
        }
    """
    def __init__(self, *args, **kwds):
        """
        Processes require a name (string) and a names (list)
        """
        AttrDict.__init__(self, *args, **kwds)
        
        has_prc_name = self.has_key('name')
        has_prc_names = self.has_key('names')
        
        self.setdefault('exclude', False)
        
        if not (has_prc_name and has_prc_names):
            raise KeyError, "ProcessGroupDict must be initialized with name and names"

    def __add__(self,y):
        """
        Process1 + Process2 is a group process containing the 
        components of each
        """
        if isinstance(y,Process):
            if y.exclude:
                new_names = list(set(self.names).difference(y.names))
            else:
                new_names = list(set(self.names).union(y.names))
            new_name = self.name + '+' + y.name
        else:
            raise TypeError, "Only processes can be added together"
            
        return Process(name =  new_name, names = new_names)


def parse_and_run():
    import os
    from optparse import OptionParser
    from warnings import warn
    from permm.mechanisms import __all__ as all_mechs
    from permm.analyses import __all__ as all_analyses
    all_mechs = '|'.join(all_mechs)
    all_analyses = '|'.join(all_analyses)
    parser = OptionParser()
    parser.set_usage("Usage: python -m permm [-c %s] [-g|-i|-a %s] [mrgfile]" % (all_mechs, all_analyses))
    parser.add_option("-i", "--interactive", dest="interactive", \
                        action="store_true", default=False, \
                        help="open an interactive environment", \
                        metavar="interactive environment")
    
    parser.add_option("-g", "--gui", dest="graphical", \
                        action="store_true", default=False, \
                        help="open a graphical user interactive environment", \
                        metavar="graphical user interactive environment")
    
    parser.add_option("-c", "--mechanism", dest="mechanism", \
                      default="cb05_camx", help="Chemical mechanisms", \
                      metavar="MECHANISM")

    parser.add_option("-a", "--analysis", dest="analysis", \
                      default=None, help="Stock analysis to perform", \
                      metavar="ANALYSIS")

    parser.add_option("-o", "--output", dest="output", \
                      default=None, help="Base output path", \
                      metavar="ouput base")
    
    (options, args) = parser.parse_args()

    from permm import mechanisms, \
                            netcdf, \
                            getmech
    from permm.Shell import load_environ
    
    get_prepared_mech = getmech.get_prepared_mech
    get_pure_mech = getmech.get_pure_mech

    mech = get_pure_mech(options.mechanism)
    start_script = 0
    if len(args) > 0:
        try:
            try:
                NetCDFFile = netcdf.NetCDFFile
                mrg_file = NetCDFFile(args[0],'r')
            except:
                from PseudoNetCDF.camxfiles.Memmaps import irr
                NetCDFFile = irr
                mrg_file = NetCDFFile(args[0])

            mech.set_mrg(mrg_file)
           
            
        except (IOError, RuntimeError), (e):
            warn("\n-------------------------------------------------------\nFirst argument was not a data file.\nAttempting to continue with first argument as a script.\n-------------------------------------------------------")
        except ValueError, (e):
            warn("\n-------------------------------------------\nAre you sure this is a %s data file?\n-------------------------------------------" % options.mechanism)
            raise e
        else:
            start_script = 1

    from permm.Shell import PERMConsole
    console = PERMConsole()
    load_environ(mech,console.locals)

    for script in args[start_script:]:
        if os.path.isfile(script):
            execfile(script, globals(), console.locals)
        else:
            exec(script, globals(), console.locals)


    if options.graphical:
        console.runsource("from permm.GUI import StartGUI")
        console.runsource("StartGUI(mech)")

    if options.analysis is not None:
        if len(args)<1:
            parser.error(msg="Requires a pyPA mrg file as an argument for analysis output.  You can enter an interactive environment (-i), but will not have access to \"netted\" reactions")

        if options.analysis == "net_balance":
            console.runsource("from permm.analyses.net_balance import net_balance")
            console.runsource("net_balance('%s', '%s', '%s')" % (options.mechanism, args[0], options.output))
        elif options.analysis == "history":
            console.runsource("from permm.analyses.history import matrix")
            console.runsource("history = matrix(mech, [C2O3], [HC+Radical-OH-HO2-O1D], [])")
            console.runsource("history.run()")
        else:
            raise "Unkown analysis"

    if options.interactive:
        load_environ(mech,console.locals)
        console.interact()

        
if __name__ == '__main__':
    parse_and_run()


from pdb import set_trace
from numpy import zeros, exp, newaxis, rollaxis
from permm.SpeciesGroup import Species

class matrix(object):
    def __init__(self, mech, nodes, traceable, intermediates, bottomup = True):
        """
            mechanism - permm.Mechanism.Mechanism object
            nodes - list of species to trace
            intermediates - species that should be immediately attributed to predecessors
            bottomup - trace from node to root? no, then root to node
        """
        # Internalize arguments for later use
        self.__mech = mech
        self.__traceable = set(traceable)
        self.__intermediates = set(intermediates)
        self.__bottomup = bottomup
        
        # make sure that nodes is a list of nodes
        if isinstance(nodes, Species):
            nodes = [nodes]

        # Initialize dictionaries for storing
        # source data
        self.origins = {}
        self.origin_loss = {}

        # Producers is a dictionary (keys = species; values = dictionaries of producers) 
        # that contains integrated production arrays
        #
        # Example: {'CO': {'Transport': array([8,5,6,...]), ...}}
        self.producers = {}
        
        # Concentrations is a dictionary (keys = species; values dictionary 
        # initial, average, and final concentration arrays.  Similar to producers
        self.concentrations = {}
        
        # Production is a dictionary of gross production each species
        self.production = {}
        
        # Losses is a dictionary of gross consumption of each species
        self.losses = {}

        # Preset traced nodes as those provided by the user
        self.traced = set(nodes)
        
        
        # Storing data about input shapes
        tmp_init = mech('INIT')
        self.__old_shape = tmp_init[tmp_init.keys()[0]].shape
        self.__shape = [i for i in self.__old_shape]
        self.__ntimes = mech.mrg.variables['TFLAG'].shape[-3]
        for i, s in enumerate(self.__shape):
            if s == self.__ntimes:
                self.__shape[i] = self.__ntimes + 1
                self.__time_dim = i
        self.__shape = tuple(self.__shape)
        
        self.run()
        
    def run(self):
        if self.__bottomup:
            for node in [n for n in self.traced]:
                self.add_predecessors(node)
        else:
            for node in [n for n in self.traced]:
                self.add_successors(node)

        self.update_origin_contributions()
    
    def add_predecessors(self, node, debug = False):
        """
        Recursively find reactions that consume and/or produce chemical 
        species (node)_n and attribute production to traceable reactant 
        species (node)_{n-1}.
        """
        
        mech = self.__mech
        
        # Find all reactions that produce node
        producers = mech.find_rxns(products = node)
        
        # Find all reactions that consume
        consumers = mech.find_rxns(reactants = node)

        # Create a temporary array
        ztemp = zeros(self.__old_shape, dtype = 'd')

        try:
            Initial = mech('Initial')[node].array()
            H_Trans = mech('H_Trans')[node].array()
            V_Trans = mech('V_Trans')[node].array()
            Motion = mech('Motion')[node].array()
            Emissions = mech('Emissions')[node].array()
            Deposition = mech('Deposit')[node].array()
            
            # Initial concentrations are production terms only for the first hour
            rollaxis(Initial, self.__time_dim)[1:] = 0
            ProdH_Trans = H_Trans.copy(); ProdH_Trans[H_Trans < 0] = 0
            LossH_Trans = H_Trans.copy(); LossH_Trans[H_Trans > 0] = 0
            ProdV_Trans = V_Trans.copy(); LossV_Trans[V_Trans < 0] = 0
            LossV_Trans = V_Trans.copy(); LossV_Trans[V_Trans > 0] = 0
            ProdMotion = Motion.copy(); ProdMotion[Motion < 0] = 0
            LossMotion = Motion.copy(); LossMotion[Motion > 0] = 0
            
            # Populate producers dictionary for node
            self.producers[node.name] = dict(H_Trans = ProdH_Trans,
                                             V_Trans = ProdV_Trans,
                                             Emissions = Emissions,
                                             Motion = ProdMotion,
                                             Initial = Initial,
                                             Chemistry = ztemp.copy())
            # Populate losses dictionary for node with physical
            # losses
            total_loss = self.losses[node.name] = LossH_Trans + LossV_Trans + Deposition + LossMotion
        except (KeyError, NameError), (e):
            # Populate producers dictionary for node with zeros
            self.producers[node.name] = dict(H_Trans = ztemp.copy(),
                                             V_Trans = ztemp.copy(),
                                             Emissions = ztemp.copy(),
                                             Motion = ztemp.copy(),
                                             Chemistry = ztemp.copy(),
                                             Initial = ztemp.copy())
            # Populate losses dictionary for node with physical
            # losses
            total_loss = self.losses[node.name] = ztemp.copy()
        
        # Add initial, final, and (if available average) concentration
        try:
            self.concentrations.setdefault(node.name,{})['Initial'] = mech('Initial')[node].array()
            self.concentrations[node.name]['Final'] = mech('Final')[node].array()
            average = self.concentrations[node.name]['Average'] = 0.5 * (mech('Final')[node].array() + mech('Initial')[node].array())
        except:
            self.concentrations.setdefault(node.name,{})['Initial'] = ztemp.copy()
            self.concentrations[node.name]['Final'] = ztemp.copy()
            
                
        
        # Add chemical losses
        for rxn in consumers:
            total_loss += mech(rxn)[node]
        
        # For each producing reaction, trace traceable reactants
        for rxn in producers:
            # Get reactants of this reaction
            reactants = mech.reaction_dict[rxn].reactants()
            
            # Get total net production of node by this reaction
            produces = mech(rxn)[node].copy()
            
            # Increment chemistry process
            self.producers[node.name]['Chemistry'] += produces
            
            # Create a list of traceable species in reaction reactants
            tracers = reduce(list.__add__, [list(set(tr.names()).intersection(reactants)) for tr in self.__traceable])
            ntracers = len(tracers)
            
            # The reactions production of node, must be divided among n tracers
            if ntracers > 1:
                produces = produces/ntracers

            # Create output for clarity
            if debug:
                print mech.reaction_dict[rxn]
                print " -",

            # For each traceable species in reactants
            for tracer in tracers:
                if debug:
                    print "%s," % tracer,
                    
                # Attribute tracer production of node to tracer
                node_producers = self.producers.setdefault(node.name,{})
                if node_producers.has_key(tracer):
                    node_producers[tracer] += produces
                else:
                    node_producers[tracer] = produces
            
            if debug:
                print ""
            
            # For each traceable reactant
            for tracer in tracers:
                # Check that it has not already been traced
                if mech(tracer) not in self.traced:
                    if debug:
                        print "Tracing", tracer, "from", mech.reaction_dict[rxn]
                    
                    # Store tracer as having been traced
                    self.traced.add(mech(tracer))

                    # Trace species
                    self.add_predecessors(mech(tracer))

        self.production[node.name] = sum([prod for prod in self.producers[node.name].values()])
        
        # Total 1st order loss rate is approximated by total 
        # loss divided by average concentration
        if not self.concentrations[node.name].has_key('Average'):
            average = self.concentrations[node.name]['Average'] = eval('.5 * (Initial + Final)', None, self.concentrations[node.name])
        else:
            average = self.concentrations[node.name]['Average']

        total_loss /= average

    def update_origin_contributions(self):
        """
        
        """
        mech = self.__mech
        
        for node, producers in self.producers.iteritems():
            # Capture the origins of node by source
            origin_sources = self.origins[node] = {}
            # and the loss of origins by source
            origin_losses = self.origin_loss[node] = {}

            # Load sources with initial values
            origin_sources['Initial'] = zeros(self.__shape, dtype = 'd')
            origin_sources['Initial'][..., :-1] = self.concentrations[node]['Initial'][..., :]
            
            # Losses are not origin specific
            loss_rate = self.losses[node]
            loss_rate = rollaxis(loss_rate, self.__time_dim)
            
            for origin, production in producers.iteritems():
                # Create an array to house production fro this origin
                # store it in the self.origins dictionary
                origin_sources[origin] = sources = zeros(self.__shape, dtype = 'd')
                origin_losses[origin] = losses = zeros(self.__shape, dtype = 'd')
                
                # Creating pointer arrays with the hour dimension first
                # for ease of iteration
                sources = rollaxis(sources,self.__time_dim)
                losses = rollaxis(losses, self.__time_dim)
                production = rollaxis(production, self.__time_dim)
                
                for thishr in range(1, self.__ntimes+1):
                    ## Create pointers to loss and production rates
                    loss = loss_rate[thishr-1]
                    prod = production[thishr-1]

                    ## Create pointers to the origin arrays
                    # the last hour's origin
                    old_origin = sources[thishr-1]

                    # Debugging entry point                    
                    if origin == 'C' and False:
                        set_trace()
                    
                    # Capture production and loss terms based on
                    # an ordinary differential equation approximation
                    # See Tonnesen 1995 Dissertation
                    sources[thishr] = new_origin = prod/-loss * (1 - exp(loss)) +  old_origin * exp(loss)
                    losses[thishr] = old_origin + prod - new_origin
    
    def reattribute_origins(self):
        """
        Timestep contributions from each chemical species
        should be attributed to the fractional origin of that species
        for each timestep.
        """
        fractional_origins = {}
        
        for ti in range(self.__ntimes):
            for created_spc, origins in self.origins.keys():
                while any([_k in self.__mech.species_dict for _k in origins.keys()]):
                    for origin_key in origins.keys():
                        if origin_key in self.__mech.species_dict:
                            origin_origins = self.origins.get(origin_key, {})
                                    
                    


import sys
import wx
from permm.Shell import PERMConsole, load_environ

class WXAgg(wx.Panel):
    def __init__(self, parent, mech):
        wx.Panel.__init__(self, parent)
        self.mech = mech
        self.species = mech.species_dict.keys()
        self.species.sort()

        # A button
        self.gobutton =wx.Button(self, label="Go", pos=(200, 325))
        self.Bind(wx.EVT_BUTTON, self.OnGo, self.gobutton)

        # the combobox Control
        self.methods = ['plot_rxns', 'find_rxns', 'print_rxns', 'print_irrs', 'print_net_rxn', 'plot_proc', 'make_net_rxn']
        self.method_labels= ['Plot Reactions', 'Show Rxn Ids', 'Print Rxns', 'Print IRRs', 'Print Net Rxn', 'Process Plot', 'Print Net Reaction']
        self.method_select = wx.ComboBox(self, size=(-1, -1), choices=self.method_labels, style=wx.CB_READONLY)
        self.method_select.SetValue(self.method_labels[0])

        self.reactants_select = wx.ListBox(self, size = (-1, 200), choices = self.species, style = wx.LB_MULTIPLE)
        self.products_select = wx.ListBox(self, size = (-1, 200), choices = self.species, style = wx.LB_MULTIPLE)
 
        # Checkbox
        self.logical_and = wx.CheckBox(self, label="And?", pos=(20,180))
        self.logical_and.SetValue(True)
        
        sizer = wx.FlexGridSizer(rows=4, cols=3, hgap=10, vgap=5)
        sizer.AddGrowableCol(3)
        sizer.AddSpacer(1) # R1, C1
        sizer.Add(wx.StaticText(self, label="PERMM")) # R1, C2
        sizer.AddSpacer(1) # R1, C3
        sizer.Add(wx.StaticText(self, label="Reactants")) # R2, C1
        sizer.AddSpacer(1) # R2, C2
        sizer.Add(wx.StaticText(self, label="Products")) # R2, C3
        sizer.Add(self.reactants_select) # R3, C1
        sizer.Add(self.logical_and) # R3, C2
        sizer.Add(self.products_select) # R3, C3
        sizer.AddSpacer(self.method_select) # R4, C1
        sizer.AddSpacer(1) # R4, C2
        sizer.Add(self.gobutton) # R4, C3

        sizer1 = wx.FlexGridSizer(rows=1, cols=2, hgap=10, vgap=5)
        # A multiline TextCtrl - This is here to show how the events work in this program, don't pay too much attention to it
        self.logger = wx.TextCtrl(self, size=(400,300), style=wx.TE_MULTILINE | wx.TE_READONLY)
        sizer1.Add(sizer)
        sizer1.Add(self.logger)
        
        self.SetSizer(sizer1)
        self.Fit()

        



    def GetReactants(self):
        ids = self.reactants_select.GetSelections()
        spcs = [self.species[id] for id in ids]
        return spcs

    def GetProducts(self):
        ids = self.products_select.GetSelections()
        spcs = [self.species[id] for id in ids]
        return spcs
    
    def GetMethod(self):
        return self.methods[self.method_labels.index(self.method_select.GetValue())]

    def OnGo(self,event):
        method = self.GetMethod()
        logical_and = self.logical_and.IsChecked()
        reactants = ', '.join(self.GetReactants())
        products = ', '.join(self.GetProducts())
        eval_str = '%s(reactants = [%s], products = [%s], logical_and = %s)' % (method, reactants, products, logical_and)
        self.logger.AppendText(eval_str + ':\n')
        return_text = self.mech(eval_str)
        if return_text is not None:
            print return_text

def StartWx(mech):
    app = wx.App(True)
    frame = wx.Frame(None)
    panel = WXAgg(frame, mech)
    frame.Fit()
    frame.Show()
    app.MainLoop()

if __name__ == '__main__':
    from permm.mechanisms import cb05_cmaq
    mech = cb05_cmaq()
    StartWx(mech)


from permm.ReactionGroup import Stoic
from warnings import warn

def NetTable(rxn_array, label='Unspecified'):
    result = '%s\n' % (label,)
    result += str(rxn_array)
    return result

def NetTables(mech):
    result = ''
    nrxn_labels = mech.nreaction_dict.keys()
    nrxn_labels.sort()
    for nrxn_label in nrxn_labels:
        nrxn = mech.nreaction_dict[nrxn_label]
        result += NetTable(nrxn, nrxn_label)
        result += '\n'
    
    return result
    


__all__ = ['cb05_camx_prep']

def cb05_camx_prep(mech_object):
    from permm.Mechanism import Mechanism
    from permm.SpeciesGroup import Species

    globals().update(mech_object.species_dict)
    
    ### Decorate mechanism with pseudo-Species
    # RO2 place holder
    mech_object.add_spc_to_reactions(mech_object.species_dict['RO2'])    
    
    # Secondary aldehydes
    mech_object.add_prd_to_reactions(Species(mech_object.species_dict['FORM'], name = 'sFORM'))
    mech_object.add_prd_to_reactions(Species(mech_object.species_dict['ALD'], name = 'sALD'))
    mech_object.add_prd_to_reactions(Species(mech_object.species_dict['ALD_'], name = 'sALD_'))
    mech_object.add_prd_to_reactions(Species(mech_object.species_dict['ALD2'], name = 'sALD2'))
    mech_object.add_prd_to_reactions(Species(mech_object.species_dict['ALDX'], name = 'sALDX'))
    mech_object.add_prd_to_reactions(Species(mech_object.species_dict['MGLY'], name = 'sMGLY'))
    mech_object.species_dict['sFORM'] = Species(names = ['sFORM'], stoic = [1.], name = 'sFORM')
    mech_object.species_dict['sALD'] = Species(names = ['sALD'], stoic = [1.], name = 'sALD')
    mech_object.species_dict['sALD_'] = Species(names = ['sALD_'], stoic = [1.], name = 'sALD_')
    mech_object.species_dict['sALD2'] = Species(names = ['sALD2'], stoic = [1.], name = 'sALD2')
    mech_object.species_dict['sALDX'] = Species(names = ['sALDX'], stoic = [1.], name = 'sALDX')
    mech_object.species_dict['sMGLY'] = Species(names = ['sMGLY'], stoic = [1.], name = 'sMGLY')

    # Add xHO2 when it is only a fraction of HO2 is secondary
    xHO2 = Species(name = 'xHO2', names = ['XO2'], stoic = [1.], exclude = False)
    cond = lambda x: x.has_prd(XO2) and x.has_prd(HO2) and x.get(XO2,0) <= x.get(HO2, 0)
    mech_object.add_spc_to_reactions_with_condition(xHO2,cond)
    
    # Add xHO2 when all HO2 is secondary
    xHO2 = Species(name = 'xHO2', names = ['HO2'], stoic = [1.], exclude = False)
    cond = lambda x: x.has_prd(XO2) and x.has_prd(HO2) and x.get(XO2,0) > x.get(HO2, 0)
    mech_object.add_spc_to_reactions_with_condition(xHO2,cond)
    
    # Add xHO2 when all HO2 is and HO2 production is explicit
    cond = lambda x: x.has_rct(XO2_N_TO2) and x.has_prd(HO2)
    mech_object.add_spc_to_reactions_with_condition(xHO2,cond)
    
    # Add primar HO2
    pHO2 = Species(name = 'pHO2', names = ['HO2', 'xHO2'], stoic = [1., -1])
    mech_object.add_prd_to_reactions(pHO2)
    
    # Add consumed HO2
    lHO2 = Species(name = 'lHO2', names = ['HO2'], stoic = [1.])
    mech_object.add_rct_to_reactions(lHO2)

    # Over-write default instance of lHO2
    lHO2 = Species(name = 'lHO2', names = ['lHO2'], stoic = [1.])
    mech_object.species_dict['lHO2'] = lHO2

    # Over-write default instance of pHO2
    pHO2 = Species(name = 'pHO2', names = ['pHO2'], stoic = [1.])
    mech_object.species_dict['pHO2'] = pHO2

    # Over-write default instance of xHO2
    xHO2 = Species(name = 'xHO2', names = ['xHO2'], stoic = [1.])
    mech_object.species_dict['xHO2'] = xHO2

    return mech_object




__all__ = ['SumTable']
from numpy import ndarray, where, nan

def sum_line(label, hourly, aggregate = None):
    if aggregate is None:
        aggregate = hourly.sum()
    nhrs = hourly.shape[0]
    hourly = hourly.tolist()
    
    template = '%-21s' + '%11.4f' * (nhrs + 1) + '\n'
    values = tuple([label] + hourly + [aggregate])
    
    return template % values

def time_line(hourly):
    nhrs = hourly.shape[0]
    hourly = hourly.tolist()
    
    template = '%-21s' % 'Time' + '% 11i' * nhrs + '%11s\n' % 'Sum'
    values = tuple(hourly)
    
    return template % values
    
def SumTable(mech):
    nrxns = mech.nreaction_dict
    globals().update(mech.species_dict)
    time = mech.mrg.variables['TFLAG'][:,0,1]
    def header(title):
        result = '\n%s\n' % (title,)
        result += time_line(time)
        return result
        
    result = ''
    inorg_pl_hv = nrxns['Inorganic New Radical']
    ald_pl_hv = nrxns['ORG+hv radical source']
    ox_pl_org = nrxns['Ox+org radical source']
    organic_new_rads = ald_pl_hv + ox_pl_org
    oh_pl_org = nrxns['OH+(organic+NO2)']
    total_new_rads = organic_new_rads + inorg_pl_hv
    total_prod_rads = oh_pl_org + total_new_rads
    pan_rads = nrxns['PAN Production']
    
    ### CXO3 New, Prod, and Loss
    cxo3_from_pan = pan_rads[CXO3]
    cxo3_non_pan = nrxns['CXO3 -> Radical'][CXO3]
    no2_from_cxo3 = nrxns['CXO3 -> Radical'][NO2]
    xo2_from_cxo3 = nrxns['CXO3 -> Radical'][XO2]
    ho2_from_cxo3 = nrxns['CXO3 -> Radical'][pHO2]
    no2_per_cxo3 = no2_from_cxo3 / total_prod_rads[CXO3]
    no2_per_cxo3_agg = no2_from_cxo3.sum() / total_prod_rads[CXO3].sum()
    xo2_per_cxo3 = xo2_from_cxo3 / total_prod_rads[CXO3]
    xo2_per_cxo3_agg = xo2_from_cxo3.sum() / total_prod_rads[CXO3].sum()
    ho2_per_cxo3 = ho2_from_cxo3 / total_prod_rads[CXO3]
    ho2_per_cxo3_agg = ho2_from_cxo3.sum() / total_prod_rads[CXO3].sum()
    prod_cxo3 = total_prod_rads[CXO3]
    
    result += header('CXO3 New, Prod, and Loss')
    # No photolysis production
    #result += sum_line('  Ald+hv', ald_pl_hv[CXO3])
    result += sum_line('  Ox+org (inc: NO3)', ox_pl_org[CXO3])
    result += sum_line('Total new CXO3', total_new_rads[CXO3])
    result += sum_line('OH+org', oh_pl_org[CXO3])
    result += sum_line('Total CXO3 Prod', prod_cxo3)
    result += sum_line('CXO3+NO2 PAN Prod', cxo3_from_pan)
    result += sum_line('CXO3+NO Loss', cxo3_non_pan)
    result += sum_line('  NO2', no2_from_cxo3)
    result += sum_line('  XO2', xo2_from_cxo3)
    result += sum_line('  HO2', ho2_from_cxo3)
    result += sum_line('Total CXO3 Loss', cxo3_from_pan + cxo3_non_pan)    
    result += sum_line('NO2/CXO3', no2_per_cxo3, no2_per_cxo3_agg)
    result += sum_line('XO2/CXO3', xo2_per_cxo3, xo2_per_cxo3_agg)
    result += sum_line('HO2/CXO3', ho2_per_cxo3, ho2_per_cxo3_agg)

    ### C2O3 New, Prod, and Loss
    c2o3_from_pan = pan_rads[C2O3]
    c2o3_non_pan = nrxns['C2O3 -> Radical'][C2O3]
    no2_from_c2o3 = nrxns['C2O3 -> Radical'][NO2]
    meo2_from_c2o3 = nrxns['C2O3 -> Radical'][MEO2]
    no2_per_c2o3 = no2_from_c2o3 / total_prod_rads[C2O3]
    no2_per_c2o3_agg = no2_from_c2o3.sum() / total_prod_rads[C2O3].sum()
    meo2_per_c2o3 = meo2_from_c2o3 / total_prod_rads[C2O3]
    meo2_per_c2o3_agg = meo2_from_c2o3.sum() / total_prod_rads[C2O3].sum()
    prod_c2o3 = total_prod_rads[C2O3]
    
    result += header('C2O3 New, Prod, and Loss')
    result += sum_line('  Ald+hv', ald_pl_hv[C2O3])
    result += sum_line('  Ox+org (inc: NO3)', ox_pl_org[C2O3])
    result += sum_line('Total new C2O3', total_new_rads[C2O3])
    result += sum_line('OH+org', oh_pl_org[C2O3])
    result += sum_line('Total C2O3 Prod', prod_c2o3)
    result += sum_line('C2O3+NO2 PAN Prod', c2o3_from_pan)
    result += sum_line('C2O3+NO Loss', c2o3_non_pan)
    result += sum_line('  NO2', no2_from_c2o3)
    result += sum_line('  MEO2', meo2_from_c2o3)
    result += sum_line('  -OOX', nrxns['C2O3 -> Radical'][RO2])
    result += sum_line('Total C2O3 Loss', c2o3_from_pan + c2o3_non_pan)
    result += sum_line('NO2/C2O3', no2_per_c2o3, no2_per_c2o3_agg)
    result += sum_line('MEO2/C2O3', meo2_per_c2o3, meo2_per_c2o3_agg)

    ### MEO2 New, Prod, and Loss
    prod_meo2 = meo2_from_c2o3+total_prod_rads[MEO2]
    ho2_from_meo2 = nrxns['MEO2 -> Radical'][pHO2]
    no2_from_meo2 = nrxns['MEO2 -> Radical'][NO2]
    no2_per_meo2 = no2_from_meo2 / prod_meo2
    no2_per_meo2_agg = no2_from_meo2.sum() / prod_meo2.sum()
    ho2_per_meo2 = ho2_from_meo2 / prod_meo2
    ho2_per_meo2_agg = ho2_from_meo2.sum() / prod_meo2.sum()

    result += header('MEO2 New, Prod, and Loss')
    result += sum_line('  Ald+hv', ald_pl_hv[MEO2])
    result += sum_line('Total new MEO2', total_new_rads[MEO2])
    result += sum_line('MEO2 from C2O3', meo2_from_c2o3)
    result += sum_line('OH+org', oh_pl_org[MEO2])
    result += sum_line('Total MEO2 Prod', prod_meo2)
    result += sum_line('MEO2+NO Loss', nrxns['MEO2 -> Radical'][MEO2])
    result += sum_line('  HO2', ho2_from_meo2)
    result += sum_line('  NO2', no2_from_meo2)
    result += sum_line('NO2/MEO2', no2_per_meo2, no2_per_meo2_agg)
    result += sum_line('HO2/MEO2', ho2_per_meo2, ho2_per_meo2_agg)
    
    ### XO2 New, Prod, and Loss
    prod_xo2 = total_prod_rads[XO2_TO2] + xo2_from_cxo3
    prod_xo2n = total_prod_rads[XO2N]
    prod_xo2_n = prod_xo2 + prod_xo2n
    ho2_from_xo2 = nrxns['Whole_Mechanism'][xHO2]
    no2_per_xo2_n = hourly = nrxns['XO2_N_TO2 -> Radical'][NO2] / prod_xo2_n
    no2_per_xo2_n_agg = agg    = nrxns['XO2_N_TO2 -> Radical'][NO2].sum() / prod_xo2_n.sum()
    ho2_per_xo2_n = hourly = ho2_from_xo2 / prod_xo2_n
    ho2_per_xo2_n_agg =  ho2_from_xo2.sum() / prod_xo2_n.sum()

    result += header('XO2 New, Prod, and Loss')
    result += sum_line('  Ald+hv', ald_pl_hv[XO2_TO2])
    result += sum_line('  Ox+org (inc: NO3)', ox_pl_org[XO2_TO2])
    result += sum_line('Total new XO2', total_new_rads[XO2_TO2])
    result += sum_line('XO2 from CXO3', xo2_from_cxo3)
    result += sum_line('OH+org', oh_pl_org[XO2_TO2])
    result += sum_line('Total XO2 Prod', prod_xo2)
    result += sum_line('  Ox+org (inc: NO3)', ox_pl_org[XO2N])
    result += sum_line('Total new XO2N', organic_new_rads[XO2N])
    result += sum_line('OH+org', oh_pl_org[XO2N])
    result += sum_line('Total XO2N Prod', prod_xo2n)
    result += sum_line('Total XO2/XO2N Prod', prod_xo2_n)
    result += sum_line('  XO2+NO Loss', nrxns['XO2_TO2 -> Radical'][XO2_TO2])
    result += sum_line('  XO2N+NO Loss', nrxns['XO2N -> Radical'][XO2N])
    result += sum_line('Total XO2/XO2N Loss', nrxns['XO2_N_TO2 -> Radical'][XO2_N_TO2])
    result += sum_line('  NO', nrxns['XO2_N_TO2 -> Radical'][NO])
    result += sum_line('  NO2', nrxns['XO2_N_TO2 -> Radical'][NO2])
    result += sum_line('  HO2', ho2_from_xo2)
    result += sum_line('  NTR', nrxns['XO2_N_TO2 -> Radical'][NTR])
    result += sum_line('  -OOX', nrxns['XO2_N_TO2 -> Radical'][RO2])
    result += sum_line('NO2/XO2', no2_per_xo2_n, no2_per_xo2_n_agg)
    result += sum_line('HO2/XO2', ho2_per_xo2_n, ho2_per_xo2_n_agg)

    ### HO2 New, Prod, and Loss
    ho2_from_hco3 = nrxns['HCO3 -> Radical'][HO2]
    prod_ho2 = total_prod_rads[pHO2] + ho2_from_meo2 + ho2_from_xo2 + ho2_from_hco3
    no2_from_ho2 = nrxns['HO2 -> Radical'][NO2]
    oh_from_ho2 = nrxns['HO2 -> Radical'][OH]
    no2_per_ho2 = no2_from_ho2 / prod_ho2
    no2_per_ho2_agg = no2_from_ho2.sum() / prod_ho2.sum()
    oh_per_ho2 = oh_from_ho2 / prod_ho2
    oh_per_ho2_agg = oh_from_ho2.sum() / prod_ho2.sum()

    result += header('HO2 New, Prod, and Loss')
    result += sum_line('Inorg+hv', inorg_pl_hv[pHO2])
    result += sum_line('  Ald+hv', ald_pl_hv[pHO2])
    result += sum_line('  Ox+org (inc: NO3)', ox_pl_org[pHO2])
    result += sum_line('Total new HO2', total_new_rads[pHO2])
    result += sum_line('HO2 from MEO2', ho2_from_meo2)
    result += sum_line('HO2 from XO2', ho2_from_xo2)
    result += sum_line('HO2 from HCO3', ho2_from_hco3)
    result += sum_line('OH+org', oh_pl_org[pHO2])
    result += sum_line('Total HO2 Prod', prod_ho2)
    result += sum_line('  CXO3', nrxns['RXN_108'][lHO2])
    result += sum_line('  O3', nrxns['RXN_13'][lHO2])
    result += sum_line('  CRO', nrxns['RXN_137'][lHO2])
    result += sum_line('  HO2', nrxns['RXN_34'][lHO2])
    result += sum_line('  H2O', nrxns['RXN_35'][lHO2])
    result += sum_line('  OH', nrxns['RXN_43'][lHO2])
    result += sum_line('  O', nrxns['RXN_44'][lHO2])
    result += sum_line('  NO3', nrxns['RXN_48'][lHO2])
    result += sum_line('  XO2/XO2N', (nrxns['RXN_56']+nrxns['RXN_57'])[lHO2])
    result += sum_line('  MEO2', nrxns['RXN_69'][lHO2])
#    result += sum_line('  FORM', nrxns['RXN_79'][lHO2])
#    result += sum_line('  HCO3', nrxns['RXN_82'][lHO2])
    result += sum_line('  C2O3', nrxns['RXN_92'][lHO2])
    result += sum_line('  NO', nrxns['RXN_30'][lHO2])
    result += sum_line('    NO2', no2_from_ho2)
    result += sum_line('    OH', oh_from_ho2)
    result += sum_line('Total HO2 Loss', nrxns['Whole_Mechanism'][lHO2])
    result += sum_line('NO2/HO2',no2_per_ho2,no2_per_ho2_agg)
    result += sum_line('OH/HO2',oh_per_ho2,oh_per_ho2_agg)

    ### OH New, Prod, and Loss
    oh_per_cxo3 = ( \
                    xo2_per_cxo3 * ho2_per_xo2_n + \
                    ho2_per_cxo3 \
                   ) * \
                   oh_per_ho2
    oh_from_new_cxo3 = total_new_rads[CXO3] * oh_per_cxo3
    oh_from_reused_cxo3 = (prod_cxo3 - total_new_rads[CXO3]) * oh_per_cxo3
                   
    oh_per_c2o3 = meo2_per_c2o3 * ho2_per_meo2 * oh_per_ho2
    oh_from_new_c2o3 = total_new_rads[C2O3] * oh_per_c2o3
    oh_from_reused_c2o3 = (prod_c2o3 - total_new_rads[C2O3]) * oh_per_c2o3

    oh_per_meo2 = ho2_per_meo2 * oh_per_ho2
    oh_from_new_meo2 = total_new_rads[MEO2] * oh_per_meo2
    oh_from_reused_meo2 = (prod_meo2 - total_new_rads[MEO2]) * oh_per_meo2
    
    oh_per_xo2 = ho2_per_xo2_n * oh_per_ho2
    oh_from_new_xo2_n = total_new_rads[XO2_N_TO2] * oh_per_xo2
    oh_from_reused_xo2_n = (prod_xo2_n - total_new_rads[XO2_N_TO2]) * oh_per_xo2

    oh_from_new_ho2 = total_new_rads[pHO2] * oh_per_ho2
    oh_from_reused_ho2 = (prod_ho2 - total_new_rads[pHO2]) * oh_per_ho2

    oh_from_new_rad = oh_from_new_cxo3 + oh_from_new_c2o3 + oh_from_new_xo2_n + \
                      oh_from_new_meo2 + oh_from_new_ho2
    oh_new_inorg = inorg_pl_hv[OH]
    
    oh_total_new = oh_from_new_rad + oh_new_inorg
    
    oh_total = nrxns['Total OH Reacted'][OH]
    oh_from_reused_rad = -oh_total - oh_total_new
    
    oh_chain_length = -oh_total / oh_total_new
    oh_chain_length_agg = -oh_total.sum() / oh_total_new.sum()
    
    oh_propef = 1 - 1/oh_chain_length
    oh_propef_agg = 1 - 1/oh_chain_length_agg

    oh_reacted_with_co = -mech.make_net_rxn([OH,CO],[])[OH]

    oh_reacted_with_voc = -mech.make_net_rxn([OH,VOC],[])[OH]
    
    oh_reacted_with_hc = -mech.make_net_rxn([OH, HC],[])[OH]

    oh_reacted_with_no2 = -mech.make_net_rxn([OH,NO2],[])[OH]

    result += header('OH New, Prod, and Loss')
    result += sum_line('  nOH from CXO3', oh_from_new_cxo3)
    result += sum_line('  nOH from C2O3', oh_from_new_c2o3)
    result += sum_line('  nOH from XO2', oh_from_new_xo2_n)
    result += sum_line('  nOH from MEO2', oh_from_new_meo2)
    result += sum_line('  nOH from HO2', oh_from_new_ho2)
    result += sum_line('nOH from new Rad', oh_from_new_rad)
    result += sum_line('nOH from Inorg+hv', oh_new_inorg)
    result += sum_line('nOH Total', oh_total_new)
    result += sum_line('OH reused Rad (diff)', oh_from_reused_rad)
    result += sum_line('OH Total Reacted', oh_total)
    result += sum_line('OH Chain length', oh_chain_length, oh_chain_length_agg)
    result += sum_line('OH Prop Eff', oh_propef, oh_propef_agg)
    result += sum_line('OH + CO', oh_reacted_with_co)
    result += sum_line('OH + VOC', oh_reacted_with_voc)
    result += sum_line('OH + HC', oh_reacted_with_hc)
    result += sum_line('OH + NO2', oh_reacted_with_no2)

    ### NO2 New, Prod, and Loss
    no2_emissions = mech('Emissions[NO2].array()')
    no2_htrans = mech('H_Trans[NO2].array()')
    no2_htrans_gain = where(no2_htrans > 0, no2_htrans, 0)
    no2_htrans_loss = where(no2_htrans < 0, no2_htrans, 0)
    no2_vtrans = mech('V_Trans[NO2].array()')
    no2_vtrans_gain = where(no2_vtrans > 0, no2_vtrans, 0)
    no2_vtrans_loss = where(no2_vtrans < 0, no2_vtrans, 0)
    no2_motion = mech('Motion[NO2].array()')
    no2_motion_gain = where(no2_motion > 0, no2_motion, 0)
    no2_motion_loss = where(no2_motion < 0, no2_motion, 0)
  
    no2_initial = mech('Initial[NO2]').array()
    no2_carry  = mech('Initial[NO2]').array()
    no2_initial[1:] = 0
    no2_carry[:1] = 0

    no2_new = no2_initial + no2_emissions + no2_htrans_gain + \
              no2_vtrans_gain + no2_motion_gain

    result += header('New NO2, Srcs Outside Box')
    result += sum_line('  NO2 initial', no2_initial)
    result += sum_line('  NO2 emissions', no2_emissions)
    result += sum_line('  NO2 horiz trans', no2_htrans_gain)
    result += sum_line('  NO2 vert trans', no2_vtrans_gain)
    result += sum_line('  NO2 entrain/dil', no2_motion_gain)
    result += sum_line('New NO2', no2_new)

    no2_phot = nrxns['NO2 Photolysis1']
    no2_from_o3_oxid = where(no2_phot[O3]<0, no2_phot[NO2], 0)
    no2_from_ro2_oxid = nrxns['NO+RO2 Oxidation'][NO2]
    no2_from_no3_pl_org = nrxns['NO3+org radical source'][NO2]
    
    no2_available = no2_initial + no2_carry + no2_from_o3_oxid + no2_new + no2_from_ro2_oxid


    # Available[NO2] - CHEM[NO2] - PHY[NO2] = 0
    # only if no2_available includes no2_initial
    no2_available_agg = (no2_initial + no2_from_o3_oxid + no2_new + no2_from_ro2_oxid).sum()
    
    no2_from_chem = no2_from_ro2_oxid + no2_from_no3_pl_org
    no2_frac_from_chem = no2_from_chem / no2_available
    
    no2_frac_from_chem[0] = no2_from_chem[0] / (no2_available[0] - no2_initial[0])

    # Skipping assumptions
    # Assuming initial NO2 chemicaly produced fraction is 
    # proportional to first hour process fractions
    #no2_from_chem[0] += no2_frac_from_chem[0] * no2_initial[0]

    no2_frac_from_chem_agg = no2_from_chem.sum() / no2_available_agg

    for hr in range(no2_from_chem.shape[0])[1:]:
        no2_from_chem[hr] += no2_frac_from_chem[hr-1] * no2_carry[hr]
        no2_frac_from_chem = no2_from_chem / no2_available
    result += header('Total Available NO2 Inside Box')
    result += sum_line('Initial', no2_initial)
    result += sum_line('Carry over', no2_carry, 0)
    result += sum_line('O3+NO Oxid', no2_from_o3_oxid)
    result += sum_line('New NO2', no2_new)
    result += sum_line('(RO2+Ox)+NO Oxid',no2_from_ro2_oxid + no2_from_no3_pl_org)
    result += sum_line('Total Available NO2', no2_available, no2_available_agg)
    result += sum_line('Frac NO2 from chem', no2_frac_from_chem, no2_frac_from_chem_agg)
    
    no2_final = -mech('Final[NO2]').array()
    no2_dep = mech('Deposit[NO2]').array()
    
    no2_phy_loss = no2_final + no2_dep + no2_htrans_loss + no2_vtrans_loss + no2_motion_loss
    no2_phy_loss_agg = no2_final[-1] + (no2_dep + no2_htrans_loss + no2_vtrans_loss + no2_motion_loss).sum()
    result += header('Physical NO2 Losses')
    result += sum_line('  NO2 Final', no2_final, no2_final[-1])
    result += sum_line('  NO2 deposition', no2_dep)
    result += sum_line('  NO2 horiz trans', no2_htrans_loss)
    result += sum_line('  NO2 vert  trans', no2_vtrans_loss)
    result += sum_line('  NO2 entrain/dil', no2_motion_loss)
    result += sum_line('Total NO2 phy losses', no2_phy_loss, no2_phy_loss_agg)
    result += sum_line('Total NO2 for Chem', no2_phy_loss + no2_available, no2_phy_loss_agg + no2_available_agg)
    
    no2_phot_loss = where(no2_phot[NO2]<0, no2_phot[NO2], 0)
    no_prod = where(no2_phot[NO2]<0, no2_phot[NO], 0)
    o3_prod = where(no2_phot[NO2]<0, no2_phot[O3], 0)
    no3_prod = where(no2_phot[NO2]<0, no2_phot[NO3], 0)
    
    o3_per_no2 = o3_prod / no2_available
    o3_per_no2_agg = o3_prod.sum() / no2_available_agg

    no_per_no2 = no_prod / no2_available
    no_per_no2_agg = no_prod.sum() / no2_available_agg
    
    no2_to_pans = nrxns['PAN Production'][NO2]
    no2_to_ntr = -nrxns['NO2 Termination'][NTR]
    no2_to_hno3 = -nrxns['NO2 Termination'][HNO3]
    no2_term = no2_to_pans + no2_to_ntr + no2_to_hno3
    no_new_sec = no2_new * no_per_no2

    no_initial = mech('Initial[NO]').array()
    no_carry  = mech('Initial[NO]').array()
    no_initial[1:] = 0
    no_carry[:1] = 0
    no_emissions = mech('Emissions[NO].array()')
    no_htrans = mech('H_Trans[NO].array()')
    no_htrans_gain = where(no_htrans > 0, no_htrans, 0)
    no_htrans_loss = where(no_htrans < 0, no_htrans, 0)
    no_vtrans = mech('V_Trans[NO].array()')
    no_vtrans_gain = where(no_vtrans > 0, no_vtrans, 0)
    no_vtrans_loss = where(no_vtrans < 0, no_vtrans, 0)
    no_motion = mech('Motion[NO].array()')
    no_motion_gain = where(no_motion > 0, no_motion, 0)
    no_motion_loss = where(no_motion < 0, no_motion, 0)
    no_from_o3_oxid = where(no2_phot[O3]<0, no2_phot[NO], 0)
    
    no_total_new = no_initial + no_carry + no_from_o3_oxid + \
                   no_emissions + no_htrans_gain + \
                   no_vtrans_gain + no_motion_gain + \
                   no_new_sec

    no_total_new_agg = (no_initial +  no_from_o3_oxid + \
                   no_emissions + no_htrans_gain + \
                   no_vtrans_gain + no_motion_gain + \
                   no_new_sec).sum()

    no_chain_length = no_prod / no_total_new
    no_chain_length_agg = no_prod.sum() / no_total_new_agg

    no_recreated = no_prod - no_new_sec
    no_recreated_agg = no_prod.sum() - no_new_sec.sum()
    no_propef = 1 - 1/where(no_chain_length == 0, nan, no_chain_length)
    no_propef_agg = 1 - 1/no_chain_length_agg


    result += header('Chemical NO2 Losses')
    result += sum_line('   HNO3 formation', no2_to_hno3)
    result += sum_line('   PAN  formation', no2_to_pans)
    result += sum_line('   NTR  formation', no2_to_ntr)
    result += sum_line('Total NO2 Term Loss', no2_term)
    result += sum_line('   NO2 Photolyzed', no2_phot_loss)
    result += sum_line('   NO  Produced', no_prod)
    result += sum_line('   O3  Produced', o3_prod)
    result += sum_line('   NO3 Produced', no3_prod)
    result += sum_line('Total NO2 Chem Loss', no2_term + no2_phot_loss)
    result += sum_line('O3 yld per NO2 phot', o3_per_no2, o3_per_no2_agg)
    result += sum_line('NO yld per NO2 avail', no_per_no2, no_per_no2_agg)
    result += sum_line('new NO frm new NO2', no_new_sec)
    result += sum_line('recreated NO', no_recreated, no_recreated_agg)

    ### New NO, Physical NO sources

    result += header('New NO, Physical NO sources')  
    result += sum_line('  Initial', no_initial)
    result += sum_line('  Carry over', no_carry, 0)
    result += sum_line('  O3+NO Oxid', no_from_o3_oxid)
    result += sum_line('  NO emissions', no_emissions)
    result += sum_line('  NO horiz trans', no_htrans_gain)
    result += sum_line('  NO vert trans', no_vtrans_gain)
    result += sum_line('  NO entrain/dil', no_motion_gain)
    result += sum_line('  sec new NO', no_new_sec)
    result += sum_line('Total new NO', no_total_new, no_total_new_agg)
    result += sum_line('NO P_n', no_propef, no_propef_agg)
    result += sum_line('NO chain length', no_chain_length, no_chain_length_agg)
    
    return result



__all__ = ['irr_plot', 'phy_plot', 'phy_plots']


HeadURL="$HeadURL: /Users/barronh/Development/permm/src/permm/graphing/timeseries.py $"
ChangeDate = "$LastChangedDate: 2011/03/11 20:10:49 $"
RevisionNum= "$LastChangedRevision: 249:cd194c9476ed $"
ChangedBy  = "$LastChangedBy: barronh $"
__version__ = RevisionNum

from warnings import warn
from datetime import datetime
from types import InstanceType
from permm.netcdf import NetCDFFile
from PseudoNetCDF.sci_var import PseudoNetCDFFile
from numpy import array, concatenate, zeros, arange, ceil, concatenate
from pylab import figure, title as pylabtitle, plot_date, savefig, legend, axis, twinx, rcParams
from matplotlib.dates import DateFormatter, date2num
from matplotlib.cm import get_cmap
from matplotlib.font_manager import FontProperties
import re
import operator
import os

if rcParams['text.usetex']:
    rcParams['text.latex.preamble'] = '\usepackage[font=sf]{mhchem}'

def get_dates(mrg_file, end_date = True, slice = slice(None)):
    date_ints = mrg_file.variables['TFLAG'][...,0,:].reshape(-1, 2)
    date_objs = array([datetime.strptime("%iT%06i" % (d,t), "%Y%jT%H%M%S") for d,t in date_ints]).reshape(mrg_file.variables['TFLAG'].shape[:-2])[slice]
    if end_date:
        date_objs = concatenate([date_objs[[0]]-(date_objs[-1]-date_objs[-2]), date_objs])
    else:
        date_objs = concatenate([date_objs, date_objs[[-1]]+(date_objs[-1]-date_objs[-2])])

    return date_objs
    
def get_date_steps(mrg_file, end_date = True, slice = slice(None)):
    date_objs = get_dates(mrg_file, end_date, slice = slice)

    date_objs = date_objs.repeat(2,0)[1:-1]
    return date_objs

def none_defaults_to(v, default):
    return {True: default, False: v}[v is None]

def plot(mech, y, stepped = True, end_date = True, figure_settings = {}, axis_settings = {}, line_settings = {}):
    
    if stepped:
        date_objs = get_date_steps(mech.mrg, end_date)
        y = y.repeat(2,0)
    else:
        date_objs = get_dates(mech.mrg, end_date)

    axis_settings = axis_settings.copy()
    axis_settings.setdefault('xlabel', 'Time')
    line_settings = line_settings.copy()
    line_settings.setdefault('linestyle', '-')
    line_settings.setdefault('linewidth', 3)
    line_settings.setdefault('marker', 'None')

    fig = figure(**figure_settings)
    ax = fig.add_subplot(111, **axis_settings)
    
    ax.plot_date(date_objs, y, **line_settings)
    ax.xaxis.set_major_formatter(DateFormatter('%jT%H'))
    fig.autofmt_xdate()

    return fig
    
def irr_plot(
             mech, reactions, species,
             factor = 1, slice = slice(None), units = None, end_date = True,
             chem = 'CHEM', title = None, ylim = None, xlim = None, 
             figure_settings = {}, axis_settings = {}, line_settings = {},
             fig = None, cmap = None, ncol = 2
            ):
    __doc__ = """
    irr_plot makes a plot with lines for each reaction object provided
    
      mech - Mechanism object
      reactions - list of reaction objects
      species - species to plot from reactions
      end_date - dates are provided for the interval end
      units - string for y-axis label (defaults to mech.irr_dict.units)
      ncol - number of columns in the legend
      fig - a previous figure to plot results on
      cmap - color map to be used
      chem - the process name of the total chemistry or None to 
             not plot total chemistry
      factor - a scaling factor for process data (i.e. if data is ppm 
               and you want to plot ppt, factor = 1e6)
      title - a string for the title of the figure
      ylim - an iterable (e.g. tuple or list) of min and max values
             for y-scale
      xlim - an iterable (e.g. tuple or list) of min and max values
             for x-scale
      figure_settings - properties for the figure; dictionary of 
                        keywords that pertain to the matplotlib
                        figure function
      axis_settings - properties for the figure axes; dictionary of 
                      keywords that pertain to the matplotlib
                      axes function
      line_settings - properties for all lines; dictionary of keywords
                      for matplotlib plot_date function
    """
    
    date_objs = get_date_steps(mech.mrg, end_date, slice = slice)
    units = none_defaults_to(units, getattr(mech.irr_dict, 'units', 'Unknown'))
    nlines = len(reactions)
    if rcParams['text.usetex']:
        title_str = 'Plot of \ce{%s} Reactions' % (species.name, )
    else:
        title_str = 'Plot of %s Reactions' % (species.name, )

    title_str = none_defaults_to(title, title_str)
    
    colors = iter(get_cmap(cmap)(arange(nlines, dtype = 'f')/(nlines-1)))
    if fig is None:
        fig = figure(**figure_settings)
        legend_line_height = .037
        if rcParams['text.usetex']: legend_line_height = .045
        nrows = ceil(nlines/ncol)
        legend_space = (nrows + 2) * legend_line_height
        ax = fig.add_axes([.1, .1 + legend_space, .8, .8 - legend_space], **axis_settings)
    else:
        ax = fig.get_axes()[-1]
        
    pylabtitle(title_str % locals())

    options = line_settings
    options.setdefault('linestyle', '-')
    options.setdefault('linewidth', 3)
    options.setdefault('marker', 'None')
    for rxn in reactions:
        options['color'] = colors.next()
        data = rxn[species] * factor
        rxn_sum = rxn.sum()
        reaction_label = rxn.display(digits = None)
        options['label'] = reaction_label
        if rcParams['text.usetex']: options['label'] = '\ce{' + options['label'] + '}'
        ax.plot_date(date_objs, data[slice].repeat(2,0), **options)

    if chem is not None:
        options['color'] = 'black'
        options['marker'] = 'o'
        options['label'] = 'Total Chem'
        try:
            data = mech('%s'  % (chem,))[species][slice].array()
        except:
            warn('Using sum of reactions for %(species)s' % locals())
            data = mech.make_net_rxn(species, species, False)[species][slice]
    
        ax.plot_date(date_objs, data.repeat(2,0), **options)

    ax.set_xlabel('Time')
    ax.set_ylabel(units)
    ax.xaxis.set_major_formatter(DateFormatter('%jT%H'))
    ax.set_xlim(date2num(date_objs[0]), date2num(date_objs[-1]))
    if ylim is not None:
        ax.set_ylim(*ylim)

    if xlim is not None:
        ax.set_xlim(*xlim)

    fig.autofmt_xdate()
    labels = [l.get_label() for l in ax.lines]
    handles = ax.lines
    leg = fig.legend(handles, labels, ncol = ncol, loc = 8, prop = FontProperties(size=10))

    return fig

def phy_plot(mech, species, init = 'INIT', final = 'FCONC', factor = 1, end_date = True,
             exclude = ['UCNV', 'AVOL', 'DATE', 'TIME', 'I', 'J', 'K'], processes = None,
             units = None, ylim = None, xlim = None,
             title = None, filter = True, fig = None, ax = None, secondaryconc = True, cmap = None,
             figure_settings = {}, axis_settings = {},
             line_settings = {'linestyle': '-', 'linewidth': 3, 'marker': 'None'},
             legend_settings = {},
             **kwds):

    """
    mech - permm.Mechanism.Mechanism object
    species - permm.SpeciesGroup.Species object
    title - title
    init - Name of initial concentration process
    final - Name of final concentration process
    ncol - number of legend columns (default: 1)
    fig - figure to plot on (default: None)
    cmap - matplotlib color map for lines (default: None)
    filter - remove processes with zero values (default: True)
    kwds - * <process name1> - process names from mech.process_dict can be 
                               provided to limit the processes shown.  When 
                               provided, a process should be a dictionary of 
                               matplotlib plot options (common: linestyle,
                               linewidth, label, marker).  The dictionary can
                               be empty.
           * <process name2> - same as process 1
           * <process nameN> - same as process 1
           * CONC - same as process 1, but for the concentration line
           * end_date - times are for the time period end
           
           
    """
    processes = none_defaults_to(processes, set(kwds.keys()).intersection(mech.process_dict.keys()))
    if processes == set():
        processes = set([k for k, v in mech.process_dict.iteritems() if len(v.names)==1 and v.names[0]==k]).difference([init, final]+exclude)

    processes = dict(zip(processes, map(lambda k: kwds.get(k,line_settings.copy()), processes)))
            
    if rcParams['text.usetex']:
        title_str = '\ce{%s} Processes' % species.name
    else:
        title_str = '%s Processes' % species.name

    title_str = none_defaults_to(title, title_str)

    if isinstance(processes, (list, set)):
        processes = dict([(k,{}) for k in processes])

    nlines = len(processes)
    colors = iter(get_cmap(cmap)(arange(nlines, dtype = 'f')/(nlines-1)))
    
    date_objs = get_date_steps(mech.mrg, end_date)
    
    if ax is None:
        if fig is None:
            fig = figure(figsize = (8,6))
            ax = fig.add_axes([0.1, 0.12, 0.8, 0.8], **axis_settings)
        else:
            try:
                ax = fig.get_axes()[0]
            except:
                ax = fig.add_axes([0.1, 0.12, 0.8, 0.8], **axis_settings)
            
    
    if isinstance(secondaryconc, bool):
        if secondaryconc:
            if len(fig.get_axes()) == 2:
                tax = fig.get_axes()[-1]
            else:
                tax = twinx(ax)
        else:
            tax = ax
    else:
        # Assuming secondaryconc is an axis
        tax = secondaryconc
        
    ax.set_title(title_str % locals())
    options = kwds.get('CONC',line_settings.copy())
    if options is not None:
        options.setdefault('color', 'k')
        options.setdefault('label', 'Conc')
        options.setdefault('linestyle', '-')
        options.setdefault('linewidth', 3)
        options.setdefault('marker', 'None')
        init_vals = mech('%s'  % (init,))[species]
        fconc_vals = mech('%s'  % (final,))[species]
        data = concatenate([init_vals[..., :1], fconc_vals[...]], axis = -1) * factor
        tax.plot_date(get_dates(mech.mrg), data, **options)
    units = kwds.get('units', None)
    for process in processes.keys():
        options = processes[process]
        options.setdefault('color', colors.next())
        options.setdefault('linestyle', '-')
        options.setdefault('linewidth', 3)
        options.setdefault('marker', 'None')
        process_label = process
        if rcParams['text.usetex']:
            process_label = process_label.replace('_', '\_')
            
        options.setdefault('label', process_label)
        var = mech('(%s)' % (process,))
        units = units or var.get_units(species)
        data = var[species].repeat(2,0) * factor
        if data.nonzero()[0].any() or not filter:
            ax.plot_date(date_objs, data, **options)
            
    if ax.get_xlabel() == '':
        ax.set_xlabel('Time')
    if ax.get_ylabel() == '':
        ax.set_ylabel('processes (%s)' % units)

    if not tax is ax:
        tax.set_ylabel('concentration')

    ax.xaxis.set_major_formatter(DateFormatter('%jT%H'))
    ax.set_xlim(date2num(date_objs[0]), date2num(date_objs[-1]))

    if ylim is not None:
        ax.set_ylim(*ylim)
    if xlim is not None:
        ax.set_xlim(*xlim)

    fig.autofmt_xdate()
    legend_settings.setdefault('prop', FontProperties(size=10))
    
    if tax == ax:
        ax.legend(**legend_settings)
    else:
        legend_lines = [l for l in tax.lines + ax.lines if l.get_label()[:5] != '_line' and l.get_label() is not None]
        legend_labels = [l.get_label() for l in legend_lines]
        ax.legend(legend_lines, legend_labels, **legend_settings)
    return fig

def _add_mech(conf):
    if conf.has_key('mech'):
        mech = conf['mech']
    else:
        from permm import get_pure_mech
        mech = conf['mech'] = get_pure_mech('_'.join([conf['mechanism'].lower(), conf['model'].lower()]))
        if isinstance(conf['mrgfile'], (PseudoNetCDFFile, InstanceType)):
            mrg_file = conf['mrgfile']
        else:
            mrg_file = NetCDFFile(conf['mrgfile'],'r')

        mech.set_mrg(mrg_file)

def phy_plots(conf, filter = True, fmt = 'pdf', fig_append = 'IPR'):
    _add_mech(conf)
    
    mech = conf['mech']
    date_objs = get_date_steps(mech.mrg, conf)

    for species, species_options in conf['species'].iteritems():
        try:
            fig = phy_plot(conf, mech,  date_objs, species, species_options, filter = filter)
            fig.savefig(os.path.join(conf['outdir'], '%s_%s.%s' % (species, fig_append, fmt)), format = fmt)
        except KeyError, detail:
            warn(detail)
    
import unittest

class TestPlots(unittest.TestCase):
    def setUp(self):
        from permm.mechanisms import small_strato
        from PseudoNetCDF import PseudoNetCDFFile
        from numpy import arange, newaxis
        from numpy.random import normal, poisson, random, seed
        
        self.mech = small_strato()
        mrg = self.mrg = PseudoNetCDFFile()
        
        mrg.createDimension('TSTEP', 10)
        mrg.createDimension('ROW', 3)
        mrg.createDimension('LAY', 4)
        mrg.createDimension('COL', 5)
        mrg.createDimension('VAR', 10)
        mrg.createDimension('DATE-TIME', 2)
        
        tflag = mrg.createVariable('TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME'))
        tflag.units = '<YYYYJJJ, HHMMSS>'
        tflag.long_name = tflag.var_desc = 'time '
        
        tflag[:, :, 0] = 2004001
        tflag[:, :, 1] = arange(10)[:, newaxis] * 10000
        
        for i in range(1, 11):
            var = mrg.createVariable('IRR_%d' % i, 'f', ('TSTEP', 'LAY', 'ROW', 'COL'))
            var.units = 'ppt'
            var.long_name = var.var_desc = 'Integrated rate for IRR ordinal %d' % i
            seed(1)
            if i % 2 == 0:
                data = arange(3 * i, 3 * i + 10)[:, newaxis, newaxis, newaxis].repeat(4, 1).repeat(3, 2).repeat(5, 3)
            elif i % 1 == 0:
                data = arange(2 * i, 2 * i + 10)[:, newaxis, newaxis, newaxis].repeat(4, 1).repeat(3, 2).repeat(5, 3)
            else:
                data = arange(1 * i, 1 * i + 10)[:, newaxis, newaxis, newaxis].repeat(4, 1).repeat(3, 2).repeat(5, 3)

            var[:] = data * {True: -1, False: 1}[i % 2 == True]

        self.mech.set_mrg(mrg)

    def testIRR(self):
        from tempfile import TemporaryFile
        tf = TemporaryFile()
        fig = self.mech("plot_rxns(O1D, O1D, False, slice = (slice(None), 0, 1, 2), show = False)")
        fig.savefig(tf)
        tf.seek(0,0)
        tfdata = tf.read()
        testdata = file('testdata/test1.png').read()
        import pdb; pdb.set_trace()
        self.assertTrue(tfdata == testdata)

if __name__ == '__main__':
    unittest.main()


def MakeMultiDiGraph(mech, traceable = None, slice = slice(None)):
    import networkx as nx
    from numpy import sum
    G = nx.MultiDiGraph()
    for spc in traceable:
        for rxn in mech.get_irrs(reactants = spc):
            rxn_slice = rxn[slice]
            rcts = [mech(rct) for rct in rxn.reactants()]
            rct_frac = rxn_slice[spc]/sum([rxn_slice[rct] for rct in rcts], axis = 0)
            for prd_name in rxn.products():
                prd = mech(prd_name)
                if prd in traceable:
                    G.add_edge(spc, prd, (rxn_slice[prd] * rct_frac).sum())
                    
    return G

def MakeDiGraph(mech, traceable = None, slice = slice(None)):
    import networkx as nx
    G = MakeMultiDiGraph(mech = mech, traceable = traceable, slice = slice)
    NewG = nx.DiGraph()
    for edge in G.edges_iter():
        u,v = edge
        vals = G[u][v]
        NewG.add_edge(u, v, sum(vals))
        
    return NewG
    
def MakeUniDiGraph(mech, traceable = None, slice = slice(None)):
    import networkx as nx
    G = MakeDiGraph(mech = mech, traceable = traceable, slice = slice)
    NewG = nx.DiGraph()
    g_edges = G.edges()
    for edge in g_edges:
        u,v = edge
        data = G.get_edge_data(u, v)
        if NewG.has_edge(v, u):
            new_data = NewG.get_edge_data(v, u) - data
            if data > 0:
                NewG[v][u] = new_data
            else:
                NewG.delete_edge(v, u)
                NewG.add_edge(u, v, abs(new_data))
        else:
            NewG.add_edge(u, v, data)
    
    return NewG

def MakeCarbonTrace(mech, slice = slice(None), traceable = None, makes_larger = []):
    from numpy.ma import masked_invalid
    traceable = [spc for spcn, spc in mech.species_dict.iteritems() if spc.atom_dict.get(spc.name, {}).get('C', 0) > 0]
    print traceable
    #if len(traceable) == 0: traceable = mech('[eval(n) for n in HC.names()]')
    import networkx as nx
    from numpy import sum
    G = nx.MultiDiGraph()
    for spc in traceable:
        spc_C = spc.atoms('C')[spc]
        for rxn in mech.get_irrs(reactants = spc):
            rxn_slice = rxn[slice].sum()
            rcts = [mech(rct) for rct in rxn.reactants()]
            for prd_name in rxn.products():
                prd = mech(prd_name)
                try:
                    prd_C = prd.atoms('C')[prd]
                except KeyError:
                    # Product has no carbon
                    continue

                if prd_C <= spc_C or spc in makes_larger:
                    if prd in traceable:
                        rct_frac = rxn_slice[spc]/sum([rxn_slice[rct] for rct in rcts if rct in traceable and rct.atoms('C')[rct] >= prd_C], axis = 0)
                        spc_portion = masked_invalid(rxn_slice[prd] * rct_frac).sum()
                        print spc, spc_C, prd, prd_C, rct_frac
                        G.add_edge(spc, prd, weight = spc_portion)
                    
    return G



__all__=['timeseries', 'barplot']
if __name__ != '__main__':
    import timeseries
    import barplot
else:
    from optparse import OptionParser
    parser = OptionParser()
    parser.set_usage("Usage: %prog [-t] [-m <model>]  <yamlfile>")
    parser.add_option("-t", "--template", dest="template",action="store_true",default=False,
                        help="Output template on standard out (configurable with -m and -c", metavar="Template")
    
    parser.add_option("-m", "--model", dest="model",default="new",
                        help="Model can either be camx, cmaq or wrfchem (for use with -t)", metavar="MODEL")
    
    (options, args) = parser.parse_args()
    if options.template:
        from os.path import join,abspath,dirname
        import sys
        cmaq_template_path = join(abspath(dirname(__file__)),'defaults','template.yaml')       
        print >> sys.stdout, file(cmaq_template_path,'r').read()
        parser.exit()
    if len(args)<1:
        parser.error(msg="Requires a yaml file as an argument.  For a template use the -t option.  The template will be output to the stdout.")
        parser.exit()
    else:
        import yaml
        from permm.graphing.timeseries import phy_plots, chem_plots, rxn_plots
        conf = yaml.load(file(args[0]))
        phy_plots(conf)
        chem_plots(conf)
        rxn_plots(conf)



__all__ = ['mz4_chem_prep']

def z4_chem_prep(mech_object):
    from permm.Mechanism import Mechanism
    from permm.SpeciesGroup import Species

    globals().update(mech_object.species_dict)
    
    return mech_object



from numpy import dtype, \
                  zeros, \
                  ndarray, \
                  array, \
                  float32, float64, int8, int16, int32, int64
import operator

ReactionGroup = str
__all__ = ['Species']

class Species(object):
    """
    Species is a class for species groups.  The simplest case
    is a single species group with a stoichiometry of 1.
    
    The simplest way to create a species is with keywords

        Species(
                name = '<name for the species>',
                names = [<species name1>, ..., <species nameN>],
                stoic = [<factor1>, ..., <factorN>],
                atom_dict = {<species name1> = dict(C=[<factor1>, ..., <factorN>],
                                                N=[<factor1>, ..., <factorN>],
                                                ...)
                         ...
                         <species nameN> = {...}
                         }
                exclude = <boolean>
               )
    
    Species may also be created by combining other species
    with a new name
    
        s1 = Species(**kwds)
        s2 = Species(**kwds)
        Species(s1, s2, name = new_name)
    
    The exclude property determines if the species will be evaluated
    as an exclusion (i.e. not species).  This property may be emulated
    with the __neg__ interface.

    Once a species has been created, it can be 
    1) indexed for stoichiometry
    2) added to other species to create larger species groups,
    3) subtracted from another species
    4) multiplied by a constant
    
    
    1) indexed for stoichiometry
        >>> NO
        Species([(1.0,)], 
              dtype=[('NO', '<f8')])
              exclude = False
        >>> NO['NO']
        array([ 1.])
    
    2) added to other species to create larger species groups,
    2a) inclusive addition
        >>> NO
        Species([(1.0,)], 
              dtype=[('NO', '<f8')])
              exclude = False
        >>> NO2
        Species([(1.0,)], 
              dtype=[('NO2', '<f8')])
              exclude = False
        >>> NOx = NO+NO2
        >>> NOx
        Species([(1.0, 1.0)], 
              dtype=[('NO2', '<f8'), ('NO', '<f8')])
              exclude = False
        
    2b) exclusive addition
        >>> xNO = Species(NO, name='xNO')
        >>> xNO.exclude = True
        >>> xNO
        Species([(1.0,)], 
              dtype=[('NO', '<f8')])
              exclude = True
        >>> NOx
        Species([(1.0, 1.0)], 
              dtype=[('NO2', '<f8'), ('NO', '<f8')])
              exclude = False
        >>> NOx+xNO
        Species([(1.0,)], 
              dtype=[('NO2', '<f8')])
              exclude = False

    2c) exclusive addition with the __neg__ interface
        >>> NOx+-NO
        Species([(1.0,)], 
              dtype=[('NO2', '<f8')])
              exclude = False

    3) subtracted from another species
        >>> NOx
        Species([(1.0, 1.0)], 
              dtype=[('NO2', '<f8'), ('NO', '<f8')])
              exclude = False
        >>> NOx-NO
        Species([(1.0,)], 
              dtype=[('NO2', '<f8')])
              exclude = False
        
    4) multiplied by a constant
        >>> OLE
        Species([(1.0,)], 
              dtype=[('OLE', '<f8')])
              exclude = False
        >>> 2*OLE
        Species([(2.0,)], 
              dtype=[('OLE', '<f8')])
              exclude = False
        >>> Species(2*OLE, name = 'OLEC')
        Species([(2.0,)], 
              dtype=[('OLE', '<f8')])
              exclude = False    
    """
    
    def __init__(self, *args, **kwds):
        if kwds.has_key('names') and kwds.has_key('stoic'):
            # n_names = len(kwds['names'])
            #
            # species_type = dtype(dict(names = kwds['names'], formats = 'd' * n_names))
            #
            # self = zeros((1,), dtype = species_type).view(type = subtype)
            #
            # for name,stoic in zip(kwds['names'],kwds['stoic']):
            #     self[name] += stoic
            self._stoic = {}
            self._stoic.update(dict(zip(kwds['names'], kwds['stoic'])))
            default_atoms = {}
            default_exclude = False
        else:
            args_are_species_groups = all(map(lambda x: isinstance(x, Species),args))

            if args == () or not args_are_species_groups:
                raise ValueError, "Species requires either names and stoic or arguments of type Species"
            else:
                if len(args) == 1:
                    result = args[0]
                else:
                    result = reduce(Species.__add__,args)
            self._stoic = result._stoic.copy()
            self.name = result.name
            default_atoms = result.atom_dict
            default_exclude = result.exclude

        self.atom_dict = kwds.pop('atom_dict', default_atoms).copy()
        try:
            self.name = kwds['name']
        except:
            # Assume that result already has a name
            pass
        self.exclude = kwds.get('exclude', default_exclude)
        self._roles = kwds.get('roles',['u']*len(self.names()))
        
    def __getitem__(self,item):
        """Test"""
        if isinstance(item,Species):
            if item.name in self.keys():
                n = item.name
                return item[n]*self[n]
            else:
                return sum([item[n]*self[n] for n in item.names()])
        else:
            result = dict.__getitem__(self._stoic,item)
            return result
            
    def __str__(self):
        if len(self.names()) == 1:
            if self[self.names()[0]] == 1. and self.name == self.names()[0]:
                return self.name
        bool_op = (' = ', ' != ')[self.exclude]
        result = self.name + bool_op + ' + '.join(['%.3f*%s' % (self[spc],spc) for spc in self.names()])
        return result
        
    def __repr__(self):
        return self.__str__()
        #try:
        #    exclude = str(self.exclude)
        #except:
        #    exclude = '?'
        #return ndarray.__repr__(self)+'\n      exclude = '+exclude
        
    def names(self):
        return [n for n in self.keys()]
    
    def __neg__(self):
        return Species(self, name = '-(%s)' % self.name, exclude = not self.exclude)
    
    def __contains__(self, lhs):
        if isinstance(lhs, Species):
            return len(set(lhs.keys()).intersection(self.keys())) > 0
        elif isinstance(lhs, str):
            return lhs in self.keys()
            
    def __rmul__(self, y):
        return self.__mul__(y)

    def __mul__(self, y):
        is_number = isinstance(y,(int,long,float, float32, float64, int8, int16, int32, int64))
        is_reaction = isinstance(y,ReactionGroup)
        
        if is_number:
            new_name_stoic = [(k, v * y) for k, v in self.iteritems()]
            new_stoic = [v for k, v in new_name_stoic]
            new_name = "%s * %f" % (self.name,float(y))
            new_names = [k for k, v in new_name_stoic]
            new_exclude = y <= 0
            return Species(name = new_name, names = new_names, stoic = new_stoic, exclude = new_exclude, atom_dict = self.atom_dict.copy())
        elif is_reaction:
            raise FutureWarning, "Working on it"
        else:
            raise TypeError, "Can only multiply species by reactions"
    
    def __sub__(self, other_species):
        return self.__add__(-other_species)
        
    def __add__(self, other_species):
        return species_sum([self, other_species])
    
    def atoms(self, atom):
        from mechanisms import atoms
        if atom in atoms:
            name_stoic = [(n, self.atom_dict[n][atom]) for n in self.names() if self.atom_dict.get(n, {}).has_key(atom)]
            if name_stoic == []:
                raise KeyError, "Atom provided (%s) is not in %s" % (atom, self.name)
            else:
                return Species(name = self.name + ':' + atom, names = [n for n, s in name_stoic], stoic = [s for n, s in name_stoic])
        else:
            raise KeyError, "Atom provided (%s) is not an atom" % atom
    
    def copy(self):
        result =  1*self
        result.name = self.name
        return result
    
    def keys(self):
        return self._stoic.keys()

    def iteritems(self):
        return self._stoic.iteritems()

def species_sum(species_list):
    if not all([isinstance(spc,Species) for spc in species_list]):
        raise TypeError, "Can only add SpeciesGroups"
        
    out_species = species_list[0].copy()
    out_names = set(out_species.keys())
    out_name = out_species.name
    out_exclude = out_species.exclude
    out_stoic = {}
    out_atoms = out_species.atom_dict.copy()
    for spc in out_species.keys():
        try:
            out_stoic[spc] += out_species[spc]
        except:
            out_stoic[spc] = float(out_species[spc])
    
    for next_species in species_list[1:]:
        if out_exclude ^ next_species.exclude:
            out_name += ' + ' + next_species.name
            out_names = out_names.difference(next_species.keys())
        else:
            out_name += ' + ' + next_species.name
            out_names.update(next_species.keys())
        
        out_exclude = out_exclude and next_species.exclude
        
        for spc in next_species.keys():
            try:
                out_stoic[spc] += next_species[spc]
            except:
                out_stoic[spc] = next_species[spc]

        if next_species.exclude:
            for spc in next_species.atom_dict.keys():
                del out_atoms[spc]
        else:
            out_atoms.update(next_species.atom_dict)
        
    out_names = list(out_names)
    out_stoic = [out_stoic[name] for name in out_names]
    return Species(name = out_name, names = out_names, \
                    stoic = out_stoic, exclude = out_exclude, atom_dict = out_atoms)

import unittest

class SpeciesTestCase(unittest.TestCase):
    def setUp(self):
        self.species = dict(OH = Species(names = ['OH'], stoic = [1], exclude = False, atom_dict = dict(OH = dict(H = 1, O = 1)), name = 'OH'),
                            HO2 = Species(names = ['HO2'], stoic = [1], exclude = False, atom_dict = dict(HO2 = dict(H = 1, O = 2)), name = 'HO2'),
                            HOx = Species(names = ['HO2', 'OH'], stoic = [1, 1], exclude = False, atom_dict = dict(HO2 = dict(H = 1, O = 2), OH = dict(H = 1, O = 1)), name = 'HOx'),
                            O3 = Species(names = ['O3'], stoic = [1], exclude = False, atom_dict = dict(O2 = dict(O = 3)), name = 'O3'),
                            )


    def assertEqualSpecies(self, s1, s2, neg = False):
        self.assertEquals(s1._stoic, s2._stoic)
        self.assertEquals(s1.atom_dict, s2.atom_dict)
        self.assertEquals(s1._roles, s2._roles)
        if neg:
            self.assertEquals(s1.exclude, not s2.exclude)
        else:
            self.assertEquals(s1.exclude, s2.exclude)
        
    def testCreate(self):
        s1 = self.species['OH']
        s2 = Species(s1, name = 'hydroxyl')
        self.assertEqualSpecies(s1, s2)
        
    def testNeg(self):
        s1 = self.species['OH']
        s2 = -s1
        self.assertEqualSpecies(s1, s2, neg = True)
    
    def testAdd(self):
        s1 = self.species['OH']
        s2 = self.species['HO2']
        s3 = self.species['HOx']
        s4 = s1 + s2
        self.assertEqualSpecies(s3, s4)

    def testSub(self):
        s1 = self.species['OH']
        s2 = self.species['HO2']
        s3 = self.species['HOx']
        s4 = s3 - s2
        self.assertEqualSpecies(s4, s1)

    def testGet(self):
        s2 = self.species['HO2']
        s3 = self.species['HOx']
        self.assertEquals(s3['HO2'], s3[s2])
    
    def testMul(self):
        s1 = self.species['OH']
        s3 = 2 * s1
        s2 = s1 * 2
        self.assertEqualSpecies(s2, s3)
        self.assertRaises(AssertionError, self.assertEqualSpecies, s1, s3)
        
    def testContains(self):
        s1 = self.species['OH']
        s2 = self.species['O3']
        s3 = self.species['HOx']
        self.assertTrue(s1 in s3)
        self.assertFalse(s2 in s3)

    def testCopy(self):
        s1 = self.species['OH']
        s2 = s1.copy()
        self.assertEqualSpecies(s1, s2)

    def testNewFromSpc(self):
        s1 = self.species['OH']
        s2 = self.species['O3']
        s3 = self.species['HOx']
        s4 = Species(s1)
        s5 = Species(s1, s2, s3)
        self.assertEqualSpecies(s1, s4)
        self.assertEqualSpecies(s1 + s2 + s3, s5)

if __name__ == '__main__':
    unittest.main()



from distutils.core import setup
try:
    from setuptools import setup
except:
    pass
import os
import sys
from warnings import warn
netcdf_pkgs = [('netCDF4', 'Dataset', 'Variable'), \
               ('netCDF3', 'Dataset', 'Variable'), \
               ('pupynere', 'netcdf_file', 'netcdf_variable')]
for pkg, reader, var in netcdf_pkgs:
    try:
        NetCDFFile = getattr(__import__(pkg, fromlist = [reader]),reader)
        define_reader = "from %s import %s as NetCDFFile" % (pkg, reader)
        define_variable = "from %s import %s as NetCDFVariable" % (pkg, var)
        netcdfpkg = [pkg]
        break
    except ImportError, e:
        warn(e.message)
else:
    warn("Did not find a 'true' NetCDFFile reader; 'true' NetCDF functionality will be disabled")
    netcdfpkg = []
    define_reader = """
class NetCDFFile(object):
    def __init__(self, *args, **kwds):
        raise ImportError('System has no valid netCDF reader; install netcdf4-python or pupynere')
"""
    define_variable = """
class NetCDFVariable(object):
    def __init__(self, *args, **kwds):
        raise ImportError('System has no valid netCDF variable; install netcdf4-python or pupynere')
"""

print >> file(os.path.join('src', 'permm', 'netcdf.py'),'wb'), """
__all__ = ['NetCDFFile']
__doc__ = \"\"\"
.. _netcdf
:mod:`netcdf` -- netcdf import point
====================================

.. module:: netcdf
   :platform: Unix, Windows
   :synopsis: Povides a single import point for a package.  If
              a user has one of many netcdf interfaces, this module
              selects it and provides it.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
\"\"\"
%s
%s
""" % (define_reader, define_variable)

def find_packages():
    import os
    packages = []
    walker = os.walk('src')
    prefix = os.path.join(os.path.curdir,'src')
    for thisdir, itsdirs, itsfiles in walker:
        if '__init__.py' in itsfiles:
            packages.append(thisdir[len(prefix)-1:])
    
    return packages
            
def find_data():
    import os
    import re
    data_pattern = re.compile(r'.*(.|_)(yaml|nc|net|irr|phy|ptb|sum|voc|txt|xls|graffle)$')
    data = []
    prefix = os.path.join(os.path.curdir,'src', 'permm')
    walker = os.walk('src')
    for thisdir, itsdirs, itsfiles in walker:
        if thisdir != os.path.join('src','permm.egg-info'):
            data.extend([os.path.join(thisdir[len(prefix)-1:],f) for f in itsfiles if data_pattern.match(f) is not None])
    
    return data

packages = find_packages()
data = find_data()

setup(name = 'permm',
      version = '1.0rc',
      author = 'Barron Henderson',
      author_email = 'barronh@gmail.com',
      maintainer = 'Barron Henderson',
      maintainer_email = 'barronh@gmail.com',
      packages = packages,
      package_dir = {'': 'src'},
      package_data = {'permm': data},
      requires = netcdfpkg + ['numpy (>=1.2)', 'yaml']
      )
